import asyncio
import math
import random
import time
from datetime import datetime
from functools import partial

import biothings.utils.es as es
import biothings.utils.inspect as btinspect
from biothings import config
from biothings.hub import INSPECTOR_CATEGORY
from biothings.hub.databuild.backend import create_backend
from biothings.utils.common import timesofar
from biothings.utils.dataload import dict_traverse
from biothings.utils.hub_db import get_source_fullname, get_src_build, get_src_dump
from biothings.utils.loggers import get_logger
from biothings.utils.manager import BaseManager
from biothings.utils.mongo import doc_feeder, id_feeder

logging = config.logger


class InspectorError(Exception):
    pass


# commong function used to call inspector
def inspect_data(backend_provider, ids, mode, pre_mapping, **kwargs):
    col = create_backend(backend_provider).target_collection
    cur = doc_feeder(col, step=len(ids), inbatch=False, query={"_id": {"$in": ids}})
    res = btinspect.inspect_docs(cur, mode=mode, pre_mapping=pre_mapping, metadata=False, auto_convert=False, **kwargs)
    return res


class InspectorManager(BaseManager):
    def __init__(self, upload_manager, build_manager, *args, **kwargs):
        super(InspectorManager, self).__init__(*args, **kwargs)
        self.upload_manager = upload_manager
        self.build_manager = build_manager
        self.logfile = None
        self.setup_log()

    def clean_stale_status(self):
        src_dump = get_src_dump()
        srcs = src_dump.find()
        for src in srcs:
            dirty = False
            jobs = src.get("inspect", {}).get("jobs", {})
            for subsrc in jobs:
                if jobs[subsrc].get("status") == "inspecting":
                    logging.warning("Found stale datasource '%s', marking inspect status as 'canceled'" % src["_id"])
                    jobs[subsrc]["status"] = "canceled"
                    dirty = True
            if dirty:
                src_dump.replace_one({"_id": src["_id"]}, src)

    def setup_log(self):
        """Setup and return a logger instance"""
        self.logger, self.logfile = get_logger("inspect")

    def get_backend_provider_info(self, data_provider):
        data_provider_type, registerer_obj, backend_provider, ups = (None,) * 4

        if data_provider[0] == "src":
            data_provider_type = "source"
            # find src_dump doc
            # is it a full source name (dot notation) ?
            fullname = get_source_fullname(data_provider[1])
            if fullname:
                # it's a dot-notation
                src_name = fullname.split(".")[0]
            else:
                # no subsource, full source name is the passed name
                src_name = data_provider[1]
                fullname = src_name
            doc = get_src_dump().find_one({"_id": src_name})  # query by main source
            if not doc:
                raise InspectorError("Can't find document associated to '%s'" % src_name)
            # get an uploader instance (used to get the data if type is "uploader"
            # but also used to update status of the datasource via register_status()
            ups = self.upload_manager[fullname]  # potentially using dot notation
            # create uploader
            registerer_obj = self.upload_manager.create_instance(ups[0])
            registerer_obj.src_doc = doc
            backend_provider = data_provider
        else:
            try:
                data_provider_type = "build"
                registerer_obj = self.build_manager.get_builder(data_provider)
                registerer_obj.src_build = get_src_build().find_one({"_id": data_provider})
                if not registerer_obj.src_build:
                    raise InspectorError("Can't find build associated to '%s'" % data_provider)
                backend_provider = data_provider
            except Exception as e:
                raise InspectorError("Unable to create backend from '%s': %s" % (repr(data_provider), e))

        return data_provider_type, registerer_obj, backend_provider, ups

    def inspect(
        self,
        data_provider,
        mode="type",
        batch_size=10000,
        limit=None,
        sample=None,
        **kwargs,
    ):
        """
        Inspect given data provider:
        - backend definition, see bt.hub.dababuild.create_backend for
          supported format), eg "merged_collection" or ("src","clinvar")
        - or callable yielding documents
        Mode:
        - "type": will inspect and report type map found in data (internal/non-standard format)
        - "mapping": will inspect and return a map compatible for later
          ElasticSearch mapping generation (see bt.utils.es.generate_es_mapping)
        - "stats": will inspect and report types + different counts found in
          data, giving a detailed overview of the volumetry of each fields and sub-fields
        - "jsonschema", same as "type" but result is formatted as json-schema standard
        - limit: when set to an integer, will inspect only x documents.
        - sample: combined with limit, for each document, if random.random() <= sample (float),
          the document is inspected. This option allows to inspect only a sample of data.
        """
        # /!\ attention: this piece of code is critical and not easy to understand...
        # Depending on the source of data to inspect, this method will create an
        # uploader or a builder. These objects don't be behave the same while they
        # pass through pickle: uploader needs to be "unprepare()"ed so it can be
        # pickled (remove some db connection, socket), while builder must *not* be
        # unprepare() because it would reset the underlying target_name (the actual
        # target collection). Also, the way results and statuses are registered is
        # different for uploader and builder...
        # So, there are lots of "if", be careful if you want to modify that code.

        data_provider_type = None  # where to register results (if possible to do so)
        registerer_obj = None  # who should register result
        t0 = time.time()
        started_at = datetime.now().astimezone()
        self.logger.info("Inspecting data with mode %s and data_provider %s" % (repr(mode), repr(data_provider)))
        if callable(data_provider):
            raise NotImplementedError("data_provider as callable untested...")
        else:
            data_provider_type, registerer_obj, backend_provider, ups = self.get_backend_provider_info(data_provider)
            if data_provider_type == "source":
                if len(ups) > 1:
                    # recursively call inspect(), collect and return corresponding tasks
                    self.logger.debug("Multiple uploaders found, running inspector for each of them: %s" % ups)
                    res = []
                    for up in ups:
                        r = self.inspect(
                            (data_provider[0], "%s" % up.name),
                            mode=mode,
                            batch_size=batch_size,
                            limit=limit,
                            sample=sample,
                            **kwargs,
                        )
                        res.append(r)
                    return res

                assert len(ups) == 1, (
                    "More than one uploader found for '%s', "
                    "not supported (yet), use main_source.source notation" % data_provider[1]
                )

        got_error = None
        try:

            async def do():
                await asyncio.sleep(0.0)
                nonlocal mode

                pinfo = {
                    "category": INSPECTOR_CATEGORY,
                    "source": "%s" % repr(data_provider),
                    "step": "",
                    "description": "",
                }
                # register begin of inspection (differ slightly depending on type)
                if data_provider_type == "source":
                    registerer_obj.register_status("inspecting", subkey="inspect")
                elif data_provider_type == "build":
                    registerer_obj.register_status("inspecting", transient=True, init=True, job={"step": "inspect"})

                self.logger.info(
                    "Running inspector on %s (type:%s,data_provider:%s)",
                    repr(data_provider),
                    data_provider_type,
                    backend_provider,
                )

                if sample is not None:
                    self.logger.info("Sample set to %s, inspect only a subset of data", sample)
                if limit is None:
                    self.logger.info("Inspecting all the documents")
                else:
                    nonlocal batch_size
                    # adjust batch_size so we inspect only "limit" docs if batch is smaller than the limit
                    if batch_size > limit:
                        batch_size = limit
                    self.logger.info("Inspecting only %s documents", limit)
                # make it pickleable
                if data_provider_type == "source":
                    # because register_obj is also used to fetch data, it has to be unprepare() for pickling
                    registerer_obj.unprepare()
                else:
                    # NOTE: do not unprepare() the builder, we'll loose the target name
                    # (it's be randomly generated again) and we won't be able to register results
                    pass

                cnt = 0
                doccnt = 0
                jobs = []
                # normalize mode param and prepare global results
                if type(mode) == str:
                    mode = [mode]

                converters, mode = btinspect.get_converters(mode)

                inspected = {}
                for m in mode:
                    inspected.setdefault(m, {})

                backend = create_backend(backend_provider).target_collection
                for ids in id_feeder(backend, batch_size=batch_size):
                    if sample is not None:
                        if random.random() > sample:
                            continue
                    cnt += 1
                    doccnt += batch_size
                    if limit and doccnt > limit:
                        break
                    pinfo["description"] = "batch #%s" % cnt

                    def batch_inspected(bnum, i, f):
                        nonlocal inspected
                        nonlocal got_error
                        nonlocal mode
                        try:
                            res = f.result()
                            for m in mode:
                                inspected[m] = btinspect.merge_record(inspected[m], res[m], m)
                        except Exception as e:
                            got_error = e
                            self.logger.error("Error while inspecting data from batch #%s: %s" % (bnum, e))
                            raise

                    pre_mapping = "mapping" in mode  # we want to generate intermediate mapping so we can merge
                    # all maps later and then generate the ES mapping from there
                    self.logger.info("Creating inspect worker for batch #%s" % cnt)
                    job = await self.job_manager.defer_to_process(
                        pinfo,
                        partial(inspect_data, backend_provider, ids, mode=mode, pre_mapping=pre_mapping, **kwargs),
                    )
                    job.add_done_callback(partial(batch_inspected, cnt, ids))
                    jobs.append(job)

                await asyncio.gather(*jobs)

                # compute metadata (they were skipped before)
                for m in mode:
                    if m == "mapping":
                        try:
                            inspected["mapping"] = es.generate_es_mapping(inspected["mapping"])
                            # metadata for mapping only once generated
                            inspected = btinspect.compute_metadata(inspected, m)
                        except es.MappingError as e:
                            inspected["mapping"] = {"pre-mapping": inspected["mapping"], "errors": e.args[1]}
                    else:
                        inspected = btinspect.compute_metadata(inspected, m)

                # just potential converters
                btinspect.run_converters(inspected, converters)

                def fully_inspected(res):
                    nonlocal got_error
                    try:
                        res = btinspect.stringify_inspect_doc(res)
                        _map = {"results": res}
                        _map["data_provider"] = repr(data_provider)
                        _map["started_at"] = started_at
                        _map["duration"] = timesofar(t0)

                        # when inspecting with "stats" mode, we can get huge number but mongo
                        # can't store more than 2^64, make sure to get rid of big nums there
                        def clean_big_nums(k, v):
                            # TODO: same with float/double? seems mongo handles more there ?
                            if isinstance(v, int) and v > 2**64:
                                return k, math.nan
                            else:
                                return k, v

                        dict_traverse(_map, clean_big_nums)
                        # register begin of inspection (differ slightly depending on type)
                        if "mapping" in mode and "errors" in res["mapping"] and "pre-mapping" in res["mapping"]:
                            registerer_obj.register_status("failed", subkey="inspect", inspect=_map)
                            got_error = InspectorError(res["mapping"]["errors"])
                        else:
                            if data_provider_type == "source":
                                registerer_obj.register_status("success", subkey="inspect", inspect=_map)
                            elif data_provider_type == "build":
                                registerer_obj.register_status(
                                    "success", job={"step": "inspect"}, build={"inspect": _map}
                                )
                    except Exception as e:
                        self.logger.exception("Error while inspecting data: %s" % e)
                        got_error = e
                        if data_provider_type == "source":
                            registerer_obj.register_status("failed", subkey="inspect", err=repr(e))
                        elif data_provider_type == "build":
                            registerer_obj.register_status("failed", job={"err": repr(e)})

                fully_inspected(inspected)
                if data_provider_type is None:
                    return
                if got_error:
                    raise got_error

            task = asyncio.ensure_future(do())
            return task
        except Exception as e:
            self.logger.error("Error while inspecting '%s': %s" % (repr(data_provider), e))
            raise

    def flatten(self, data_provider, mode=("type", "stats"), do_validate=True):
        if isinstance(mode, tuple):
            mode = list(mode)  # may not be necessary, but previous mode default is a list, so let's be consistent
        elif isinstance(mode, str):
            mode = [mode]

        data_provider_type, registerer_obj, backend_provider, _ = self.get_backend_provider_info(data_provider)

        results = {}
        if data_provider_type == "source":
            src_sources_inspect_data = registerer_obj.src_doc.get("inspect", {}).get("jobs", {})
            for src_source_name, inspect_data in src_sources_inspect_data.items():
                results[src_source_name] = {
                    _mode: btinspect.flatten_and_validate(
                        inspect_data["inspect"]["results"].get(_mode) or {}, do_validate
                    )
                    for _mode in mode
                }
        else:
            inspect_data = registerer_obj.src_build.get("inspect", {}).get("results", {})
            results[backend_provider] = {
                _mode: btinspect.flatten_and_validate(inspect_data.get(_mode) or {}, do_validate) for _mode in mode
            }

        return results



import datetime
from functools import partial

from dateutil import parser as dtparser

from biothings import config as btconfig
from biothings.utils.manager import BaseManager

logger = btconfig.logger


class AutoArchiveManager(BaseManager):
    """This feature will add a new console command "auto_archive", a new api "/auto_archive".
    It is intended to allow automatically archive old build, based on configuration.

    This feature can be configurated by using AUTO_ARCHIVE_CONFIG variable, to determine:
        - schedule: how frequency a build configuration should run
        - days: how old a build should be archived

    AUTO_ARCHIVE_CONFIG = {
        "build_configuration_name": {
            "schedule: "* 0 * * *",  # run daily at 0am UTC
            "days": 3
        },
        ...
    }
    """

    DEFAULT_ARCHIVE_SCHEDULE = "* 0 * * *"  # run daily at 0am UTC

    def __init__(self, build_manager, job_manager, auto_archive_config=None, *args, **kwargs):
        super().__init__(job_manager, *args, **kwargs)

        self.build_manager = build_manager
        self.auto_archive_config = auto_archive_config or {}

    def configure(self):
        for build_config_name in self.build_manager.register.keys():
            archive_config = self.auto_archive_config.get(build_config_name)

            if not isinstance(archive_config, dict):
                logger.info(f"Build config: {build_config_name}: No archive config found!")
                continue

            schedule = archive_config.get("schedule") or self.DEFAULT_ARCHIVE_SCHEDULE
            days = archive_config.get("days")

            if not days or days < 0:
                logger.warning(f"Build config: {build_config_name}: Days is invalid or not give!")
                continue

            self.job_manager.submit(
                partial(self.archive, build_config_name, days=days, dryrun=False),
                schedule=schedule,
            )

    def archive(self, build_config_name, days, dryrun=True):
        """
        Archive any builds which build date is older than today's date
        by "days" day.
        """
        if not days or days < 0:
            logger.warning("Days is invalid or not give, so nothing to do!")
            return

        logger.info("Auto-archive builds older than %s days" % days)
        builds = self.build_manager.list_merge(build_config_name)
        today = datetime.datetime.now().astimezone()
        at_least_one = False
        for build_id in builds:
            build = self.build_manager.build_info(build_id)
            try:
                bdate = dtparser.parse(build["_meta"]["build_date"]).astimezone()
            except KeyError:
                logger.warning('Build "{}" missing "_meta" key.'.format(build_id))
                continue
            deltadate = today - bdate
            if deltadate.days > days:
                logger.info("Archiving build %s (older than %s)" % (build_id, deltadate))
                if dryrun:
                    logger.info('This is a dryrun of "archive(%s)", no real changes were made.', build_id)
                else:
                    self.build_manager.archive_merge(build_id)
                at_least_one = True
        if not at_least_one:
            logger.info("Nothing to archive")



import random
from itertools import chain

import tornado.escape
import tornado.httpserver
import tornado.ioloop
import tornado.locks
import tornado.options
import tornado.web
from rich import print as rprint
from rich.console import Console
from rich.panel import Panel

from biothings.utils.common import traverse
from biothings.utils.serializer import load_json, to_json


class NoResultError(Exception):
    pass


async def get_available_routes(db, table_space):
    """return a list available URLs/routes based on the table_space and the actual collections in the database"""
    collection_names = set(db.collection_names())
    list_routes = []
    detail_routes = []
    for table in table_space:
        if table in collection_names:
            rprint(f"[green]Found collection: [/green]{table}[green]; counting...[/green]", end="")
            tbl_cnt = db[table].count()
            rprint(f"[green]Done ([/green]{tbl_cnt}[green] documents)[/green]")
            if tbl_cnt > 0:
                list_routes.append(f"/{table}/")
                detail_routes.append(f"/{table}/([^/]+)/")
                db._count = getattr(db, "_count", {})
                db._count[table] = tbl_cnt  # save table count for later use
                if tbl_cnt > 100000:
                    rprint(
                        f"[yellow]WARNING: collection [/yellow]{table}[yellow] has more than 100,000 documents. \n[/yellow]"
                        f'[yellow]Queries below can be slow since the data are not indexed, except the "_id" field.[/yellow]'
                    )
    return list_routes, detail_routes


class BaseHandler(tornado.web.RequestHandler):
    def set_default_headers(self):
        self.set_header("Content-Type", "application/json")


class HomeHandler(BaseHandler):
    """the handler for the landing page, which lists all available routes"""

    async def get(self):
        list_routes, detail_routes = await get_available_routes(self.application.db, self.application.table_space)
        self.write(to_json(list_routes + detail_routes))


class DocHandler(BaseHandler):
    """The handler for the detail view of a document, e.g. /<source>/<doc_id/"""

    async def get(self, slug, item_id):
        src_cols = self.application.db[slug]
        doc = src_cols.find_one({"_id": item_id})
        if not doc:
            raise tornado.web.HTTPError(404)
        self.write(to_json(doc))


class QueryHandler(BaseHandler):
    """The handler for return a list of docs matching the query terms passed to "q" parameter e.g. /<source>/?q=<query>"""

    async def get(self, slug):
        src_cols = self.application.db[slug]

        start = self.get_argument("from", 0, True)
        limit = self.get_argument("size", 10, True)
        query_string = self.get_argument("q", "", True)
        query_params = {
            key_value.split(":", 1)[0]
            .strip()
            .strip('"')
            .strip("'"): key_value.split(":", 1)[1]
            .strip()
            .strip('"')
            .strip("'")
            for key_value in query_string.split("AND")
            if key_value and len(key_value.split(":", 1)) == 2
        }
        if limit:
            limit = int(limit)
            start = int(start)
            # entries, total_hit = src_cols.find_with_count(query_params, start=start, limit=limit)
            entries, total_hit = src_cols.findv2(
                query_params, start=start, limit=limit, return_total=True, return_list=True
            )
        else:
            # entries, total_hit = src_cols.find_with_count(query_params)
            entries, total_hit = src_cols.findv2(query_params, return_total=True)
        if not entries:
            entries = []

        self.write(
            to_json(
                {
                    "from": start,
                    "end": start + len(entries),
                    "total_hit": total_hit,
                    "entries": entries,
                }
            )
        )


def get_example_queries(db, table_space):
    """Populate example queries for a given table_space"""
    out = {}
    for table in table_space:
        col = db[table]
        total_cnt = getattr(db, "_count", {}).get(table, col.count())
        n = 5
        i = random.randint(0, min(1000, total_cnt - n))
        random_docs = [
            load_json(row[0])
            for row in (
                col.get_conn()
                .execute(
                    # f"SELECT document FROM {table} WHERE _id IN (SELECT _id FROM {table} ORDER BY RANDOM() LIMIT 10)"
                    f"SELECT document FROM {table} LIMIT {n} OFFSET {i}"
                )
                .fetchall()
            )
        ]
        key_value_list = list(chain(*[traverse(doc, leaf_node=True) for doc in random_docs]))
        selected_fields = []
        while len(selected_fields) < n:
            key, value = random.choice(key_value_list)
            if key == "_id" or not value or (isinstance(value, str) and (len(value) > 50 or " " in value)):
                continue
            selected_fields.append((key, value))
        out[table] = {"ids": [doc["_id"] for doc in random_docs], "fields": selected_fields}
    return out


class Application(tornado.web.Application):
    """The main application class, which defines the routes and handlers."""

    def __init__(self, db, table_space, **settings):
        self.db = db
        self.table_space = table_space
        handlers = [
            (r"/?", HomeHandler),
            (r"/([^/]+)/?", QueryHandler),
            (r"/([^/]+)/([^/]+)/?", DocHandler),
        ]
        settings.update({"debug": True})
        super().__init__(handlers, **settings)


async def main(host, port, db, table_space):
    """The main function, which starts the server."""
    list_routes, detail_routes = await get_available_routes(db, table_space)
    del detail_routes
    if not list_routes:
        rprint('[red]Error: Source data do not exist or are empty. Was "upload" runned successfully yet?[/red]')
        return

    app = Application(db, table_space, **{"static_path": "static"})
    app.listen(port, address=host)

    rprint(f"[green]Listening on http://{host}:{port}[/green]")
    rprint(f"[green]View all available routes: http://{host}:{port}/[/green]")
    example_queries = get_example_queries(db, table_space)
    console = Console()
    for route in list_routes:
        route = route.strip("/")
        example_ids = example_queries[route]["ids"]
        example_fields = [(k, str(v)) for k, v in example_queries[route]["fields"]]
        console.print(
            Panel(
                "\n"
                + ":link: Get a document by id:\n"
                + f"    [green]http://{host}:{port}/{route}/<doc_id>[/green]\n"
                + "    [green]Examples:[/green]\n"
                + f"     [green]http://{host}:{port}/{route}/{example_ids[0]}[/green]\n"
                + f"     [green]http://{host}:{port}/{route}/{example_ids[1]}[/green]\n"
                + ":link: Query documents by fields:\n"
                + f"    [green]http://{host}:{port}/{route}?q=<query>[/green]\n"
                + "    [green]Examples:[/green]\n"
                + f"     [green]http://{host}:{port}/{route}?from=0&size=10[/green]\n"
                + f"     [green]http://{host}:{port}/{route}?q={':'.join(example_fields[0])}[/green]\n"
                + f"     [green]http://{host}:{port}/{route}?q={':'.join(example_fields[1])} AND {':'.join(example_fields[-1])}[/green]\n",
                title=f"[bold]http://{host}:{port}/{route}[/bold]",
                title_align="left",
            )
        )

    shutdown_event = tornado.locks.Event()
    await shutdown_event.wait()



import certifi
import orjson
from tornado.httpclient import HTTPRequest

from biothings.web.analytics.events import Event, Message


class Channel:
    def handles(self, event):
        raise NotImplementedError()

    def send(self, event):
        raise NotImplementedError()


class SlackChannel(Channel):
    def __init__(self, hook_urls):
        self.hooks = hook_urls

    def handles(self, event):
        return isinstance(event, Message)

    def send(self, message):
        for url in self.hooks:
            yield HTTPRequest(
                url=url,
                method="POST",
                headers={"content-type": "application/json"},
                body=orjson.dumps(message.to_slack_payload()).decode(),
                ca_certs=certifi.where(),  # for Windows compatibility
            )


# Measurement Protocol (Universal Analytics)
# https://developers.google.com/analytics/devguides/collection/protocol/v1/devguide


class GAChannel(Channel):
    def __init__(self, tracking_id, uid_version=1):
        self.tracking_id = tracking_id
        self.uid_version = uid_version

    def handles(self, event):
        return isinstance(event, Event)

    def send(self, payload):
        events = payload.to_GA_payload(self.tracking_id, self.uid_version)
        # #batch-limitations section of the URL above
        # A maximum of 20 hits can be specified per request.
        for i in range(0, len(events), 20):
            yield HTTPRequest(
                "http://www.google-analytics.com/batch",
                method="POST",
                body="\n".join(events[i : i + 20]),  # noqa: E203
            )


class GA4Channel(Channel):
    def __init__(self, measurement_id, api_secret, uid_version=1):
        self.measurement_id = measurement_id
        self.api_secret = api_secret
        self.uid_version = uid_version

    def handles(self, event):
        return isinstance(event, Event)

    def send(self, payload):
        """

        Limitations:
        https://developers.google.com/analytics/devguides/collection/protocol/ga4/sending-events?client_type=gtag
        """
        events = payload.to_GA4_payload(self.measurement_id, self.uid_version)
        # #batch-limitations section of the URL above
        # A maximum of 25 hits can be specified per request.
        url = f"https://www.google-analytics.com/mp/collect?measurement_id={self.measurement_id}&api_secret={self.api_secret}"
        for i in range(0, len(events), 25):
            data = {
                "client_id": str(payload._cid(self.uid_version)),
                "user_id": str(payload._cid(1)),
                "events": events[i : i + 25],  # noqa: E203
            }
            yield HTTPRequest(url, method="POST", body=orjson.dumps(data))



from typing import MutableMapping, Optional, Type


class _BaseMetaClass(type):
    subclasses: MutableMapping[str, Type["_ChildContext"]] = {}

    def __new__(cls, clsname, bases, attrs):
        # produce the method that creates child contexts
        def mk_ch_context_meth(child_context, field):
            def fn(self, **kwargs):
                ch_context = cls.subclasses[child_context](self)
                for at, v in kwargs.items():
                    if not at.startswith("_") and at in dir(ch_context):
                        meth = getattr(ch_context, at)
                        if callable(meth):
                            meth(v)
                            continue
                    raise AttributeError(f"Child does not have {at}")
                self.document[field] = ch_context.document
                return ch_context

            return fn

        # produce the method that updates attributes
        def mk_attr_meth(field):
            def fn(self, v):
                self.document[field] = v
                return self

            return fn

        if "CHILD_CONTEXTS" in attrs:
            for name, (ch_context, field) in attrs["CHILD_CONTEXTS"].items():
                # FIXME: Hopefully we want the docstrings for params here
                #  Which is kind of a problem because the "more root" types
                #  are defined earlier, and to obtain the possible params
                #  we need to look at the "less root" class which is yet to be
                #  defined.
                docstring = f"""Set {field}
                            Create {ch_context} and set {field}
                            """
                meth = mk_ch_context_meth(ch_context, field)
                meth.__doc__ = docstring
                meth.__name__ = name
                attrs[name] = meth
        if "ATTRIBUTE_FIELDS" in attrs:
            for name, field in attrs["ATTRIBUTE_FIELDS"].items():
                docstring = f"""Set {field} field"""
                meth = mk_attr_meth(field)
                meth.__doc__ = docstring
                meth.__name__ = name
                attrs[name] = meth

        new_cls = type.__new__(cls, clsname, bases, attrs)
        cls.subclasses[clsname] = new_cls
        return new_cls


class _BaseContext(metaclass=_BaseMetaClass):
    # {'method_name': ('ContextClass', 'fieldName')}
    CHILD_CONTEXTS = {}
    # {'method_name: 'fieldName'}
    ATTRIBUTE_FIELDS = {}
    EXTENSION = False

    def __init__(self):
        self.document = {}

    def __getattr__(self, item: str):
        if self.EXTENSION and item.startswith("x_") and len(item) > 2:
            default_field_name = item.replace("_", "-")

            def set_extension_value(__value, field: str = default_field_name):
                if not field.startswith("x-") or len(field) < 3:
                    raise ValueError(f"Invalid extension field '{field}'")
                self.document[field] = __value
                return self

            set_extension_value.__name__ = f"{item}"
            set_extension_value.__doc__ = "Set extension"

            return set_extension_value

        raise AttributeError(f"'{self.__class__.__name__}' has no attribute '{item}'")


class _ChildContext(_BaseContext):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent

    def end(self):
        """Explicitly return to the parent context

        Returns:
            Parent context
        """
        return self.parent

    def __getattr__(self, attr_name: str):
        try:
            return super(_ChildContext, self).__getattr__(attr_name)
        except AttributeError:
            pass  # would have returned if it's an extension
        multilevel_allow = ["path", "get", "put", "post", "delete", "options", "head", "patch", "trace"]
        if not attr_name.startswith("_"):
            if attr_name in dir(self.parent) or (attr_name in multilevel_allow and hasattr(self.parent, attr_name)):
                return getattr(self.parent, attr_name)
        raise AttributeError()


class _HasParameters(_BaseContext):
    def parameter(self, name, in_, required):
        parameters = self.document.setdefault("parameters", [])
        param_context = OpenAPIParameterContext(self, name, in_, required)
        parameters.append(param_context.document)
        return param_context


class _HasSummary(_BaseContext):
    ATTRIBUTE_FIELDS = {"summary": "summary"}


class _HasDescription(_BaseContext):
    ATTRIBUTE_FIELDS = {"description": "description"}


class _HasExternalDocs(_BaseContext):
    CHILD_CONTEXTS = {
        "external_docs": ("OpenAPIExternalDocsContext", "externalDocs"),
    }


class _HasTags(_BaseContext):
    def tag(self, t):
        tags = self.document.setdefault("tags", [])
        tags.append(t)
        return self


class OpenAPIContext(_HasExternalDocs):
    CHILD_CONTEXTS = {
        "info": ("OpenAPIInfoContext", "info"),
    }
    EXTENSION = True

    def __init__(self):
        super(OpenAPIContext, self).__init__()
        self.document = {
            "openapi": "3.0.3",
            "paths": {},
        }

    def server(self, url: str, description: Optional[str] = None):
        servers = self.document.setdefault("servers", [])
        server = {"url": url}
        if description:
            server["description"] = description
        servers.append(server)
        return self

    def path(self, path: str, summary: Optional[str] = None, description: Optional[str] = None):
        path_item = OpenAPIPathItemContext(self)
        if summary:
            path_item.summary(summary)
        if description:
            path_item.description(description)
        self.document["paths"][path] = path_item.document
        return path_item


class OpenAPIInfoContext(_ChildContext, _HasDescription):
    ATTRIBUTE_FIELDS = {
        "title": "title",
        "terms_of_service": "termsOfService",
        "version": "version",
    }
    CHILD_CONTEXTS = {
        "contact": ("OpenAPIContactContext", "contact"),
        "license": ("OpenAPILicenseContext", "license"),
    }
    EXTENSION = True


class OpenAPIContactContext(_ChildContext):
    ATTRIBUTE_FIELDS = {
        "name": "name",
        "url": "url",
        "email": "email",
    }
    EXTENSION = True


class OpenAPILicenseContext(_ChildContext):
    ATTRIBUTE_FIELDS = {
        "name": "name",
        "url": "url",
    }
    EXTENSION = True


class OpenAPIPathItemContext(_ChildContext, _HasSummary, _HasDescription, _HasParameters):
    CHILD_CONTEXTS = {}
    for http_method in ["get", "put", "post", "delete", "options", "head", "patch", "trace"]:
        CHILD_CONTEXTS[http_method] = ("OpenAPIOperation", http_method)
    EXTENSION = True


class OpenAPIOperation(_ChildContext, _HasSummary, _HasExternalDocs, _HasTags, _HasDescription, _HasParameters):
    ATTRIBUTE_FIELDS = {"operation_id": "operationId"}
    EXTENSION = True

    def __init__(self, parent):
        super(OpenAPIOperation, self).__init__(parent)
        self.document = {
            "responses": {
                "200": {
                    "description": "Success",
                }
            }
        }


class OpenAPIParameterContext(_ChildContext, _HasDescription):
    ATTRIBUTE_FIELDS = {
        "deprecated": "deprecated",
        "allow_empty": "allowEmptyValue",
        "style": "style",
        "explode": "explode",
        "allow_reserved": "allowReserved",
        "schema": "schema",
    }
    EXTENSION = True

    def __init__(self, parent, name: str, in_: str, required: bool):
        super(OpenAPIParameterContext, self).__init__(parent)
        self.document = {
            "name": name,
            "in": in_,
            "required": required,
        }

    def type(self, typ: str, **kwargs):
        schema = {
            "type": typ,
        }
        for k in ["minimum", "maximum", "default"]:
            v = kwargs.get(k)
            if v is not None:
                schema[k] = v
        self.schema(schema)
        return self


class OpenAPIExternalDocsContext(_ChildContext, _HasDescription):
    ATTRIBUTE_FIELDS = {
        "url": "url",
    }
    EXTENSION = True


OpenAPIDocumentBuilder = OpenAPIContext



import gzip
from os import listdir
from os.path import isdir, isfile, join
from tempfile import TemporaryDirectory, mkstemp

from tornado.template import Template
from tornado.web import RequestHandler, StaticFileHandler

from biothings.utils.serializer import to_json

catalog = Template(
    """
    <html>
        <head>
            <title> Biothings Hub Logs </title>
        </head>
        <body>
            <h1> Logs </h1>
            <ul>
            {% for log in logs %}
                <li>
                    <a href={{"./" + log }}>
                        {{ log[:-4] if log.endswith('.log') else log }}
                    </a>
                </li>
            {% end %}
            </ul>
        </body>
    </html>
"""
)

logfile = Template(
    """
    <html>
        <head>
            <title> {{ "Log " + name }} </title>
        </head>
        <body>
            <h1> {{ name }} </h1>
            <p>
                {% for line in lines %}
                    {{ line }} </br>
                {% end %}
            </p>
        </body>
    </html>
"""
)


def get_log_content(file_path, **kwargs):
    lines = None
    if file_path.endswith(".gz"):
        with gzip.open(file_path, "rb") as f:
            lines = f.read().decode().splitlines()
    else:
        with open(file_path, "r") as file:
            lines = file.readlines()

    if lines and "lines" in kwargs:
        cap_lines = kwargs["lines"]
        cap_lines = cap_lines[0] if isinstance(cap_lines, list) else cap_lines
        try:
            cap_lines = int(cap_lines)
            lines = lines[-cap_lines:]
            if len(lines) == cap_lines:
                lines.append(f"\n***Logs were capped at {cap_lines} lines***")
        except Exception:
            pass
    return lines


class DefaultCORSHeaderMixin:
    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Methods", "GET,OPTIONS")
        self.set_header("Access-Control-Allow-Headers", "*")


class HubLogDirHandler(DefaultCORSHeaderMixin, RequestHandler):
    def initialize(self, path):
        self.path = path

    def get(self, filename):
        fullname = join(self.path, filename)
        if isdir(fullname):
            logs = [f"{f}/" if isdir(join(fullname, f)) else f for f in listdir(fullname)]

            if "filter" in self.request.arguments:
                filters = self.get_argument("filter") or ""
                filters = filters.split(",")
                logs = set([f for keyword in filters for f in logs if keyword in f])

            logs = sorted(logs)

            if "json" in self.request.arguments:
                self.finish(to_json(list(logs)))
            else:
                self.finish(catalog.generate(logs=logs))
            return

        if not isfile(fullname):
            self.set_status(404)
            return

        lines = get_log_content(fullname, **self.request.arguments)
        self.finish(logfile.generate(name=filename, lines=lines))


class HubLogFileHandler(DefaultCORSHeaderMixin, StaticFileHandler):
    def options(self, *args, **kwargs):
        self.set_status(204)

    async def get(self, path: str, include_body: bool = True) -> None:
        """If request path is a gz file, we will uncompress it first, then return get with the uncompress file path"""

        self.path = self.parse_url_path(path)
        absolute_path = self.get_absolute_path(self.root, self.path)
        self.absolute_path = self.validate_absolute_path(self.root, absolute_path)
        if self.absolute_path is None:
            return

        if include_body:
            download_file_name = self.path.replace(".gz", "")
            self.set_header("Content-Type", "text")
            self.set_header("Content-Disposition", f"attachment; filename={download_file_name}")
            with TemporaryDirectory(dir=self.root) as temp_dir:
                _, temp_file = mkstemp(dir=temp_dir)
                lines = get_log_content(self.absolute_path, **self.request.arguments)
                lines = "\n".join(lines)
                with open(temp_file, mode="w") as fwrite:
                    fwrite.write(lines)
                return await super().get(temp_file, include_body=True)

        return await super().get(path, include_body=include_body)



"""
###########################################
# THE DEFAULT HUB CONFIGURATION VARIABLES #
###########################################

This file is used to define all Hub features’ settings with their metadata, documentation, and default value.

In your own application, you should create a config.py file by copying from the config.py.example,
then define/override the variables to fit your needs.

The documentation for a variable can be defined by one of these case:
    1. the configuration value is a ConfigurationDefault instance (specify a default value)
        or a ConfigurationError instance, in which case the documentation is taken from the instance doc.
    2. the documentation can be specified as an inline comment
    3. the documentation can be specified as comments above

There are several special comment formats used to organize and manager params:
    - all comments above a param are considered as documentation/description for the parameter,
      until a empty line or a non-comment line is found.
    - to be considered as documentation, comments also need to have a space after "#". So:
            # my comment
      will be kepts as documentation, but:
            #my comment
      will just be ignored (so python code can be commented that way, without then being part
      of the documentation)
    - A section title can be added to organize parameters. A section title looks like this:
            #* my section title *#
      It can be added for each parameters, or the first section found above is the section
      the parameter will be associated to. An empty section title can be used to reset the
      section title and associate the current parameter to default one:
            #* *#
      If no section is found, all parameters are part of the default one (None).
    - some parameters needs to be kept secret (like passwords for instance):
            #- invisible -#
      will hide the parameter, including the name, value, description, from the configuration
            #- hidden -#
      will keep the parameter in the configuration displayed to users, but its value will be omitted
            #- readonly -#
      will allow the parameter to shown, but not editable

      Note: special comments can stacked:
            #- readonly -#
            #- hidden -#
      will make the parameter read-only, and its value won't be displayed

For example:

- Defined variable using ConfigurationDefault
```
DATA_HUB_DB_DATABASE = ConfigurationDefault(
    default='biothings_hubdb',
    desc="db containing the following (internal use)"
)

or use a python statement:

DIFF_PATH = ConfigurationDefault(
        default=ConfigurationValue('set_default_folder(DATA_ARCHIVE_ROOT,"diff")'),
        desc="Define path to folder which will contain output files from diff")
```

- Defined variable using ConfigurationError
```
DATA_SRC_SERVER = ConfigurationError("Define hostname for source database")
```

- Defined variable with documentation as a comment line, and metadata
```
# API port
#- hide -#
#- readonly -#
HUB_API_PORT = 7080
```

"""

import importlib

# logging is needed for evaluating default logger value below
import logging  # pylint: disable=unused-import      # noqa
import os

import biothings.utils.jsondiff

# set_default_folder is needed for evaluating some default values below
from biothings.utils.configuration import set_default_folder  # pylint: disable=unused-import      # noqa
from biothings.utils.configuration import ConfigurationDefault, ConfigurationError, ConfigurationValue

# * 1. General *#
# Hub name/icon url/version, for display purpose
HUB_NAME = "Biothings Hub"
HUB_ICON = None

# SSH port for hub console
# - readonly -#
HUB_SSH_PORT = 7022
# API port
# - readonly -#
HUB_API_PORT = 7080
# Readonly API port
# - readonly -#
READONLY_HUB_API_PORT = ConfigurationDefault(
    default=ConfigurationValue("HUB_API_PORT+1"), desc='Readonly API port when "readonly" feature is enabled'
)

# The format is a dictionary of 'username': 'cryptedpassword'
# Generate crypted passwords with 'openssl passwd -crypt'
# - hide -#
# - readonly -#
HUB_PASSWD = {"guest": "9RKfd8gDuNf0Q"}

# Webhook to publish notifications to a Slack channel
SLACK_WEBHOOK = None

# When code changes in plugins or "manual" datasources, Hub automatically restarts
# to reflect those changes
USE_RELOADER = True

# * 2. Datasources *#
ACTIVE_DATASOURCES = ConfigurationDefault(default=[], desc="List of package paths for active datasources")

# Skip all scheduled dumper jobs
SKIP_DUMPER_SCHEDULE = False

# Skip all scheduled uploader jobs after a success dump
SKIP_UPLOADER_POLL = False

# Auto archive feature will use this configuration to get schedule config for corresponding build configuration.
# If not set it will use the default value defined in AutoArchiveManager
AUTO_ARCHIVE_CONFIG = None
"""
AUTO_ARCHIVE_CONFIG = {
    "build_configuration_name": {
        "schedule: "* 0 * * *",  # run daily at 0am UTC
        "days": 3
    },
    ...
}
"""

# Docker connection configuration
# client_url should match the following formats:
# ssh://ubuntu@remote_ip:port
# unix://var/run/docker.sock
# http://remote_ip:port
# https://remote_ip:port
DOCKER_CONFIG = {
    "docker1": {"tls_cert_path": None, "tls_key_path": None, "client_url": ""},
    "localhost": {"client_url": "unix://var/run/docker.sock"},
}

# Optional, can used to override the docker connection name from the src_url
DOCKER_HOST = None

# * 3. Folders *#
# Path to a folder to store all downloaded files, logs, caches, etc...
DATA_ARCHIVE_ROOT = ConfigurationError(
    "Define path to folder which will contain all downloaded data, cache files, etc..."
)

# where to store info about processes launched by the hub
RUN_DIR = ConfigurationDefault(
    default=ConfigurationValue("""set_default_folder(DATA_ARCHIVE_ROOT,"run")"""),
    desc="Define path to folder which will store info about processes launched by the hub",
)

# cached data (it None, caches won't be used at all)
CACHE_FOLDER = ConfigurationDefault(
    default=ConfigurationValue("""set_default_folder(DATA_ARCHIVE_ROOT,"cache")"""),
    desc="Define path to folder which will contain cache files, set to None to disable",
)

# Path to a folder to store all 3rd party parsers, dumpers, etc...
DATA_PLUGIN_FOLDER = ConfigurationDefault(
    default="./plugins", desc="Define path to folder which will contain all 3rd party parsers, dumpers, etc..."
)

# Path to folder containing diff files
# Usually inside DATA_ARCHIVE_ROOT
DIFF_PATH = ConfigurationDefault(
    default=ConfigurationValue("""set_default_folder(DATA_ARCHIVE_ROOT,"diff")"""),
    desc="Define path to folder which will contain output files from diff",
)

# Path to folder containing release note files
# Usually inside DATA_ARCHIVE_ROOT
RELEASE_PATH = ConfigurationDefault(
    default=ConfigurationValue("""set_default_folder(DATA_ARCHIVE_ROOT,"release")"""),
    desc="Define path to folder which will contain release files",
)


# Define a folder to store uploaded files via biothings.hub.api.handlder.UploadHandler
# It's enabled with "upload" feature for HubServer
DATA_UPLOAD_FOLDER = ConfigurationDefault(
    default=ConfigurationValue("""set_default_folder(DATA_ARCHIVE_ROOT,"dataupload")"""),
    desc="Define path to folder which will store uploaded files via upload handler",
)


# Root folder containing ElasticSearch backups, created
# by snapshots with repo type "fs". This setting must match
# elasticsearch.yml value, param "path.repo"
# If using "fs" type repository with post-step "archive",
# this folder must have permissions set for user/group running the hub
ES_BACKUPS_FOLDER = ConfigurationDefault(
    default=ConfigurationValue('set_default_folder(DATA_ARCHIVE_ROOT,"es_backups")'),
    desc="Define path to folder which will contain containing ElasticSearch snapshot backups",
)

# this dir must be created manually
# Usually inside DATA_ARCHIVE_ROOT
LOG_FOLDER = ConfigurationDefault(
    default=ConfigurationValue('set_default_folder(DATA_ARCHIVE_ROOT,"logs")'),
    desc="Define path to folder which will contain log files",
)

# hub logger name
LOGGER_NAME = "hub"

logger = ConfigurationDefault(
    default=ConfigurationValue("""logging.getLogger(LOGGER_NAME)"""),
    desc="Provide a default hub logger instance (use setup_default_log(name,log_folder)",
)

# * 4. Index & Diff *#
# Pre-prod/test ES definitions
INDEX_CONFIG = {
    "indexer_select": {
        # default
    },
    "env": {
        "localhub": {
            "host": "localhost:9200",
            "indexer": {"args": {"timeout": 300, "retry_on_timeout": True, "max_retries": 10}},
        }
    },
}

# Snapshot environment configuration
SNAPSHOT_CONFIG = {}

# Auto snapshot cleaner feature will use this configuration to get schedule config for corresponding environment.
AUTO_SNAPSHOT_CLEANUP_CONFIG = None
"""
AUTO_SNAPSHOT_CLEANUP_CONFIG = {
        "environment_name": {
            "schedule": "* 0 * * *",      # run daily at 0am UTC
            "keep": 3,                   # the number of most recent snapshots to keep in one group
            "group_by": "build_config",  # the attr of which its values form groups
            "extra_filters": {}          # a set of criterions to limit which snapshots are to be cleaned
        },
        ...
    }
"""

# reporting diff results, number of IDs to consider (to avoid too much mem usage)
MAX_REPORTED_IDS = 1000
# for diff updates, number of IDs randomly picked as examples when rendering the report
MAX_RANDOMLY_PICKED = 10
# size of a diff file when in memory (used when merged/reduced)
MAX_DIFF_SIZE = 50 * 1024**2  # 50MiB (~1MiB on disk when compressed)


# * 5. Release *#
# Release configuration
# Each root keys define a release environment (test, prod, ...)
RELEASE_CONFIG = {
    "env": {
        "tests3": {
            "cloud": {
                "type": "aws",  # default, only one supported by now
                "access_key": None,
                "secret_key": None,
            },
            "release": {
                "bucket": None,
                "region": None,
                "folder": "%(build_config.name)s",
                "auto": True,  # automatically generate release-note ?
            },
            "diff": {
                "bucket": None,
                "region": None,
                "folder": "%(build_config.name)s",
                "auto": True,  # automatically generate diff ? Careful if lots of changes
            },
            "publish": {
                "pre": {
                    "snapshot": [
                        {
                            "action": "archive",
                            "format": "tar.xz",
                            "name": "%(build_config.name)s_snapshot_%(_meta.build_version)s.tar.xz",
                            "es_backups_folder": ConfigurationValue("""ES_BACKUPS_FOLDER"""),
                        },
                        {
                            "action": "upload",
                            "type": "s3",
                            "bucket": ConfigurationValue("""S3_SNAPSHOT_BUCKET"""),
                            "region": ConfigurationValue("""S3_REGION"""),
                            "base_path": "%(build_config.name)s/$(Y)",
                            "file": "%(build_config.name)s_snapshot_%(_meta.build_version)s.tar.xz",
                            "acl": "private",
                            "es_backups_folder": ConfigurationValue("""ES_BACKUPS_FOLDER"""),
                            "overwrite": True,
                        },
                    ],
                    "diff": [],
                }
            },
        }
    }
}

# Specify AWS credentials to access snapshots stored in S3 bucket
STANDALONE_AWS_CREDENTIALS = {"AWS_ACCESS_KEY_ID": "", "AWS_SECRET_ACCESS_KEY": ""}

# Standalone configuration, relates to how the Hub
# install data releases. You can specify, per data release name,
# which ES host/index to address, or use the default for all data
# releases.
# Note: if data release name doesn't any key here, a _default will be
# used (it must then exist, key = "_default")
# Note: ES host can be a string which stands for a single host,
# or a dict, which contains a collection of es hosts for same index, same release name
STANDALONE_CONFIG = {
    # default config
    "_default": {
        "es_host": "localhost:9200",
        "index": "biothings_current",
    },
    # # custom definition
    # "release_name" : {
    #     "es_host" : "anotherhost:9200",
    #     "index" : "specical_index_name",
    # }
    # # release with multiple es hosts
    # "release_name_2" : {
    #     "es_host" : {
    #         "prod": "prodserver:9200",
    #         "staging": "stagingserver:9200"
    #     },
    #     "index" : "specical_index_name",
    # }
}

# Default targeted standalone version
# (once published, data is fetched and deployed by what's called
# a standalone instance). Modify thorougly (ie. don't modify it)
STANDALONE_VERSION = {"branch": "standalone_v2", "commit": None, "date": None}

# A list of URLs to the versions.json files, which contain data release metadata
VERSION_URLS = []

# Use this configuration to customize validation logic of the auto hub feature.
# the AutoHubValidator will be use as default. Any customize class must be extended from it.
AUTOHUB_VALIDATOR_CLASS = None

# Set to True to skip checking application/biothings version matching, before installing
# a data release, in version settings like "app_version", "standalone_version", "biothings_version"
SKIP_CHECK_COMPAT = True

# This configuration used for autorelease feature,
# which specifies when and how the backend should automatically install a release
# the key is the source name,
# the value can be a string that will be a croniter's schedule: Ref: https://pypi.python.org/pypi/croniter
# or it can be a dict, which has a schedule and an extra item which will be passed to the install function.
AUTO_RELEASE_CONFIG = None
# AUTO_RELEASE_CONFIG = {
#   "source_name": "*/1 * * * *",
#   "source_name": {
#     "schedule": "*/1 * * * *",
#     "extra": {
#         "force": True,
#         "use_no_downtime_method": False,
#     }
#   },
# }


# Specific how many old indices should we keep, after restore a snapshot.
# 0: delete all old index.
# 1-10: keep recent n of old indices
# -1: do not delete old index at all
RELEASE_KEEP_N_RECENT_INDICES = 0

# Specify "batch_size" passed to BaseSyncer.sync,
# which applys diffs during an incremental data release.
# Set to a lower value if the sync job puts too much load to the server.
SYNC_BATCH_SIZE = 10000

# * 6. Job Manager *#
# How much memory hub is allowed to use:
# "auto" will let hub decides (will use 50%-60% of available RAM)
# while None won't put any limits. Number of bytes can also be
# specified
HUB_MAX_MEM_USAGE = None

# Max number of *processes* hub can access to run jobs
HUB_MAX_WORKERS = int(os.cpu_count() / 4)
# Max number of *processes* used when syncing data
# (applygin diff/incremental data)
MAX_SYNC_WORKERS = HUB_MAX_WORKERS

# Max queued jobs in job manager
# this shouldn't be 0 to make sure a job is pending and ready to be processed
# at any time (avoiding job submission preparation) but also not a huge number
# as any pending job will consume some memory).
MAX_QUEUED_JOBS = os.cpu_count() * 4

# * 7. Hub Internals *#
DATA_SRC_SERVER = ConfigurationError("Define hostname for source database")
DATA_SRC_PORT = ConfigurationDefault(default=27017, desc="Define port for source database")
DATA_SRC_DATABASE = ConfigurationError("Define name for source database")
DATA_SRC_SERVER_USERNAME = ConfigurationDefault(
    default=None, desc="Define username for source database connection (or None if not needed)"
)
DATA_SRC_SERVER_PASSWORD = ConfigurationDefault(
    default=None, desc="Define password for source database connection (or None if not needed)"
)

# Target (merged collection) database connection
DATA_TARGET_SERVER = ConfigurationError("Define hostname for target database (merged collections)")
DATA_TARGET_PORT = ConfigurationDefault(default=27017, desc="Define port for source database")
DATA_TARGET_DATABASE = ConfigurationError("Define name for target database (merged collections)")
DATA_TARGET_SERVER_USERNAME = ConfigurationDefault(
    default=None, desc="Define username for target database connection (or None if not needed)"
)
DATA_TARGET_SERVER_PASSWORD = ConfigurationDefault(
    default=None, desc="Define password for target database connection (or None if not needed)"
)

DATA_HUB_DB_DATABASE = ConfigurationDefault(
    default=ConfigurationValue("""'biothings_hubdb'"""), desc="db containing the following (internal use)"
)
DATA_SRC_MASTER_COLLECTION = ConfigurationDefault(
    default=ConfigurationValue("""'src_master'"""), desc="for metadata of each src collections"
)
DATA_SRC_DUMP_COLLECTION = ConfigurationDefault(
    default=ConfigurationValue("""'src_dump'"""), desc="for src data download information"
)
DATA_SRC_BUILD_COLLECTION = ConfigurationDefault(
    default=ConfigurationValue("""'src_build'"""), desc="for src data build information"
)
DATA_PLUGIN_COLLECTION = ConfigurationDefault(
    default=ConfigurationValue("""'data_plugin'"""), desc="for data plugins information"
)
API_COLLECTION = ConfigurationDefault(
    default=ConfigurationValue("""'api'"""), desc="for api information (running under hub control)"
)
CMD_COLLECTION = ConfigurationDefault(
    default=ConfigurationValue("""'cmd'"""), desc="for launched/running commands in shell"
)
EVENT_COLLECTION = ConfigurationDefault(
    default=ConfigurationValue("""'event'"""), desc="for launched/running commands in shell"
)
HUB_CONFIG_COLLECTION = ConfigurationDefault(
    default=ConfigurationValue("""'hub_config'"""), desc="for values overrifing config files"
)
DATA_TARGET_MASTER_COLLECTION = ConfigurationDefault(default=ConfigurationValue("""'db_master'"""), desc="")

# Internal backend. Default to mongodb
# For now, other options are: mongodb, sqlite3, elasticsearch
HUB_DB_BACKEND = ConfigurationError("Define Hub DB connection")
# HUB_DB_BACKEND = {
#        "module" : "biothings.utils.sqlite3",
#        "sqlite_db_foder" : "./db",
#        }
# HUB_DB_BACKEND = {
#        "module" : "biothings.utils.mongo",
#        "uri" : "mongodb://localhost:27017",
#        #"uri" : "mongodb://user:passwd@localhost:27017", # mongodb std URI
#        }
# HUB_DB_BACKEND = {
#        "module" : "biothings.utils.es",
#        "host" : "localhost:9200",
#        }

# Hub environment (like, prod, dev, ...)
# Used to generate remote metadata file, like "latest.json", "versions.json"
# If non-empty, this constant will be used to generate those url, as a prefix
# with "-" between. So, if "dev", we'll have "dev-latest.json", etc...
# "" means production
HUB_ENV = ""

# Set whether configuration parameters can be edited
# and superseded by user through ConfigurationManger
# Note: once config manager has been configured with this
# field, it's deleted to make sure we can't change it at runtime
CONFIG_READONLY = True

# Don't check used versions, just propagate them when publishing.
# That is, Hub won't ensure that all version information is
# properly set)
SKIP_CHECK_VERSIONS = False

# cache file format ("": ascii/text uncompressed, or "gz|zip|xz"
CACHE_FORMAT = "xz"

# Deprecated. Don't it's used any more
# Role, when master, hub will publish data (updates, snapshot, etc...) that
# other instances can use (production, standalones)
BIOTHINGS_ROLE = "slave"

####################################################
# Additional optional settings avaiable if needed: #
####################################################

# Pass any optional tornado settings to tornado.httpserver.HTTPServer
# see biothings.hub.api.start_api
# TORNADO_SETTINGS = {
#    # max 10GiB upload
#    "max_buffer_size" : 10*1024*1024*1024,
# }

# don't bother with elements order in a list when diffing,
# mygene optmized uploaders can't produce different results
# when parsing data (parallelization)
importlib.reload(biothings.utils.jsondiff)
biothings.utils.jsondiff.UNORDERED_LIST = True



import asyncio
import logging
from collections import defaultdict
from datetime import datetime
from functools import reduce
from operator import add

from dateutil.parser import parse as dtparse

from biothings.utils.common import get_loop

logger = logging.getLogger(__name__)


class BiothingsMetadata:
    def __init__(self):
        # cached, generated from mappings
        self.biothing_metadata = defaultdict(dict)
        # {
        #     "variant": {
        #         "biothing_type": "variant",
        #         "build_date": "2020-05-08T17:33:59.756164",
        #         "build_version": "20200508",
        #         "src": {"cosmic": { ... }, ... },
        #         "stats": {"total": 928585791 } }
        #     "gene": { ... }
        # }
        self.biothing_mappings = defaultdict(dict)
        # {
        #     "variant": {
        #         'aeolus': {'properties': {'drug_code': {...}, ...}},
        #         'cadd': {'properties': {'1000g': {...}, 'alt': {...}, ...}},
        #         'cgi': {'properties': {'association': {...}, ...}},
        #         'chebi': {'properties': {'brand_names': {...}, ...}}, ... }
        #     "gene": { ... }
        # }
        self.biothing_licenses = defaultdict(dict)
        # {
        #     "variant": {
        #         'aeolus': 'http://bit.ly/2DIxWwF',
        #         'cadd': 'http://bit.ly/2TIuab9',
        #         'cgi': 'http://bit.ly/2FqS871',
        #         'chebi': 'http://bit.ly/2KAUCAm', ... }
        #     "gene": { ... }
        # }

    def get_metadata(self, biothing_type):  # hub
        return self.biothing_metadata[biothing_type]

    def get_mappings(self, biothing_type):
        return self.biothing_mappings[biothing_type]

    def get_licenses(self, biothing_type):
        return self.biothing_licenses[biothing_type]

    async def refresh(self, biothing_type):
        pass


class BiothingsESMetadata(BiothingsMetadata):
    def __init__(self, indices, client):
        super().__init__()

        if not indices:
            # all indices on the host
            indices = {None: "_all"}
        if None not in indices:
            # default index pattern when no type specified
            indices[None] = next(iter(indices.values()))

        self.indices = indices
        self.client = client

        # initial refresh
        loop = get_loop()
        for btype in self.indices:
            obj = self.refresh(btype)
            if asyncio.iscoroutine(obj):
                try:  # py3.8+
                    task = loop.create_task(obj, name=str(btype))
                except TypeError:
                    task = loop.create_task(obj)
                task.add_done_callback(logger.debug)

    @property
    def types(self):  # biothing_type(s)
        return tuple(filter(None, self.indices.keys()))

    def update(self, biothing_type, info, count):
        """
        Read ES index mappings for the corresponding biothing_type,
        Populate datasource info and field properties from mappings.
        """
        _type = biothing_type

        # try to resolve default to an equivalent
        # and concrete biothing_type (in meta) to display
        if _type is None:
            for type_, pattern in self.indices.items():
                if self.indices[None] == pattern:
                    _type = type_
                    break

        reader = _BiothingsESMetadataReader(_type, info, count)
        self.biothing_metadata[biothing_type] = reader.get_metadata()
        self.biothing_mappings[biothing_type] = reader.get_mappings()
        self.biothing_licenses[biothing_type] = reader.get_licenses()

    def refresh(self, biothing_type=None):
        from elasticsearch import AsyncElasticsearch, Elasticsearch

        if isinstance(self.client, Elasticsearch):
            return self._refresh(biothing_type)
        elif isinstance(self.client, AsyncElasticsearch):
            return self._async_refresh(biothing_type)

    def _refresh(self, biothing_type):
        index = self.indices[biothing_type]
        info = self.client.indices.get(index=index)
        count = self.client.count(index=index)
        self.update(biothing_type, info, count)
        return info

    async def _async_refresh(self, biothing_type):
        index = self.indices[biothing_type]
        info = await self.client.indices.get(index=index)
        count = await self.client.count(index=index)
        self.update(biothing_type, info, count)
        return info


class BiothingsMongoMetadata(BiothingsMetadata):
    def __init__(self, collections, client):
        super().__init__()

        self.collections = collections
        self.client = client

    @property
    def types(self):  # biothing_type(s)
        return tuple(filter(None, self.collections.keys()))

    async def refresh(self, biothing_type):
        collection = self.client[self.collections[biothing_type]]
        # https://pymongo.readthedocs.io/en/stable/api/pymongo/collection.html
        # #pymongo.collection.Collection.estimated_document_count
        self.biothing_metadata[biothing_type] = BiothingHubMeta(
            biothing_type=biothing_type, stats=dict(total=collection.estimated_document_count())
        ).to_dict()

    def get_mappings(self, biothing_type):
        # document database does not have data schema
        # however, it might be possible to extract those indexed fields
        # https://pymongo.readthedocs.io/en/stable/api/pymongo/collection.html
        # #pymongo.collection.Collection.list_indexes
        return {"__N/A__": True}

    def get_licenses(self, biothing_type):
        # rely on metadata storage support
        return {}


class BiothingsSQLMetadata(BiothingsMetadata):
    def __init__(self, tables, client):
        super().__init__()

        self.tables = tables
        self.client = client

    @property
    def types(self):  # biothing_type(s)
        return tuple(filter(None, self.tables.keys()))

    async def refresh(self, biothing_type):
        # https://docs.sqlalchemy.org/en/14/core/reflection.html
        # This is a temporary solution as a proof of concept.
        # The implementation should probably be refined.
        # It doesn't work with empty tables at this point.
        table = self.tables[biothing_type]
        cursor = self.client.execute(f"SELECT * FROM {table}")
        if cursor.returns_rows:
            self.biothing_mappings[biothing_type] = {
                key: {"type": type(val).__name__} for key, val in zip(cursor.keys(), cursor.fetchone())
            }
            self.biothing_metadata[biothing_type] = BiothingHubMeta(
                biothing_type=biothing_type, stats=dict(total=cursor.rowcount)
            ).to_dict()


class _BiothingsESMetadataReader:
    """
    Read http://<elasticsearch>/<index_pattern>/ and ./_stats
    If the pattern matches one index, then that index's info will be used.
    If the pattern matches multiple indices, then the results will be combined.
    If the pattern matches no index, then empty dictionaries are returned.
    """

    def __init__(self, biothing_type, info, count):
        self.biothing_type = biothing_type
        self.document_count = count

        self.indices_info = {index: _ESIndex(biothing_type, **index_info) for index, index_info in info.items()}

    def get_mappings(self):
        """
        Mapping properties used for metadata field endpoint. For example:
        {
            'aeolus': {'properties': {'drug_code': {...}, ...}},
            'cadd': {'properties': {'1000g': {...}, 'alt': {...}, ...}},
            'cgi': {'properties': {'association': {...}, ...}},
            'chebi': {'properties': {'brand_names': {...}, ...}},
            ...
        }
        """
        mappings = list(info.get_mappings() for info in self.indices_info.values())
        mappings = reduce(add, mappings).to_dict() if mappings else {}
        return mappings

    def get_licenses(self):
        """
        Source-URL pairs that contains the data licencing information. Example:
        {
            'aeolus': 'http://bit.ly/2DIxWwF',
            'cadd': 'http://bit.ly/2TIuab9',
            'cgi': 'http://bit.ly/2FqS871',
            'chebi': 'http://bit.ly/2KAUCAm',
            ...
        }
        """
        licenses = list(info.get_licenses() for info in self.indices_info.values())
        licenses = reduce(add, licenses).to_dict() if licenses else {}
        return licenses

    def get_metadata(self):
        """
        Provide description about the data under this type. Example:
        {
            "biothing_type": "variant",
            "build_date": "2020-05-08T17:33:59.756164",
            "build_version": "20200508",
            "src": {"cosmic": { ... }, ... },
            "stats": {"total": 928585791 }
        }
        """
        metadata = list(info.get_metadata() for info in self.indices_info.values())
        metadata = reduce(add, metadata).to_dict() if metadata else {}
        if metadata:
            if metadata.get("biothing_type") == "__multiple__":
                metadata["biothing_type"] = self.biothing_type
            metadata["stats"]["total"] = self.document_count["count"]
        metadata["_biothing"] = self.biothing_type
        metadata["_indices"] = list(self.indices_info.keys())
        return metadata


class _ESIndex:
    """
    Read one index's info http://<elasticsearch>/<index>.
    Return combinable BiothingMetaProp objects.
    """

    def __init__(self, biothing, aliases, mappings, settings):
        self.biothing = biothing
        self.aliases = aliases
        self.mappings = _ESIndexMappings(mappings)
        self.settings = _ESIndexSettings(settings)

    def get_metadata(self):
        """
        Return BiothingHubMetadata instance.
        Populate empty metadata basing on index settings.
        Fill in empty stats field if not provided.
        """
        if self.mappings.metadata:
            try:
                return BiothingHubMeta(**self.mappings.metadata)
            except KeyError:
                pass
        return BiothingHubMeta(
            biothing_type=self.biothing,
            build_date=self.settings.get_creation_date().isoformat(),
            build_version=self.settings.get_index_version(),
            src={},
            stats={},
        )

    def get_licenses(self):
        return BiothingLicenses(self.mappings.extract_licenses())

    def get_mappings(self):
        return BiothingMappings(self.mappings.properties)


class _ESIndexSettings:
    """
    Object representation of ES index settings.
    {
        "index": {
            "number_of_shards": "1",
            "auto_expand_replicas": "0-1",
            "provided_name": ".tasks",
            "creation_date": "1566293197607",
            "priority": "2147483647",
            "number_of_replicas": "0",
            "uuid": "yWBk0qw0QXmEuxJFas3mIg",
            "version": {
                "created": "6050099"
            }
        }
    }
    """

    def __init__(self, setting):
        self.index = setting["index"]

    def get_creation_date(self):
        return datetime.fromtimestamp(int(self.index["creation_date"]) / 1000)

    def get_index_version(self):
        if "updated" in self.index["version"]:
            return self.index["version"]["updated"]
        return self.index["version"]["created"]


class _ESIndexMappings:
    """
    Object representation of ES index mappings:
    {
        # this level is only available for es6
        "<doc_type> : {
            'properties': { ... },  ---> mapping
            '_meta': {
                "src" : { ... }     ---> licenses
                ...
            },              -----------> metadata
            ...
        }
    }
    """

    def __init__(self, mapping):
        # for elasticsearch version 6.x
        if len(mapping) == 1 and next(iter(mapping)) != "properties":
            # remove doc_type, support 1 type per index
            # mapping in ES6 is nested under doc_type
            # we will try to take the mapping dict under the first doc_type
            _mapping = next(iter(mapping.values()))
            if isinstance(_mapping, dict) and "properties" in _mapping:
                mapping = _mapping
        self.enabled = mapping.pop("enabled", True)
        self.dynamic = mapping.pop("dynamic", True)
        self.properties = mapping.get("properties", {})
        self.metadata = mapping.get("_meta", {})

    def extract_licenses(self):
        """
        Return source name - license url pairs.
        """
        licenses = {}
        for src, info in self.metadata.get("src", {}).items():
            if "license_url_short" in info:
                licenses[src] = info["license_url_short"]
            elif "license_url" in info:
                licenses[src] = info["license_url"]
        return licenses


class BiothingMetaProp:
    def __add__(self, other):
        raise NotImplementedError

        # NOTE
        # The current implementations below
        # may not be able to properly handle
        # field (key/source) collisions.

    def to_dict(self):
        raise NotImplementedError


class BiothingLicenses(BiothingMetaProp):
    def __init__(self, licenses):
        self.licenses = licenses

    def __add__(self, other):
        licenses = dict(self.licenses)
        licenses.update(other.licenses)
        return BiothingLicenses(licenses)

    def to_dict(self):
        return dict(self.licenses)


class BiothingMappings(BiothingMetaProp):
    def __init__(self, properties):
        self.properties = properties

    def __add__(self, other):
        mappings = dict(self.properties)
        mappings.update(other.properties)
        return BiothingMappings(mappings)

    def to_dict(self):
        return dict(self.properties)


class BiothingHubMeta(BiothingMetaProp):
    def __init__(self, **metadata):  # dict
        self.biothing_type = metadata.get("biothing_type")
        # self.build_date = datetime.fromisoformat(metadata['build_date']) # python3.7 syntax
        # self.build_date = datetime.strptime(metadata['build_date'], "%Y-%m-%dT%H:%M:%S.%f")
        self.build_date = metadata.get("build_date")
        self.build_version = metadata.get("build_version")
        self.src = metadata.get("src", {})
        self.stats = metadata.get("stats", {})

        if self.build_date and isinstance(self.build_date, str):
            self.build_date = dtparse(metadata["build_date"]).astimezone()

    def to_dict(self):
        return {
            "biothing_type": self.biothing_type,
            "build_date": self.build_date.isoformat() if isinstance(self.build_date, datetime) else self.build_date,
            "build_version": self.build_version,
            "src": self.src,
            "stats": self.stats,
        }

    def __add__(self, other):
        # combine biothing_type field
        biothing_type = self.biothing_type
        if other.biothing_type != self.biothing_type:
            biothing_type = "__multiple__"

        # take the latest build_date
        # TODO if one of them is None
        build_date = self.build_date
        if other.build_date and (not build_date or other.build_date > build_date):
            # set to other.build_data if it's newer or current build_date is None
            build_date = other.build_date

        # combine build_version field
        build_version = self.build_version
        if other.build_version != build_version:
            build_version = "__multiple__"

        # combine source field
        src = dict(self.src)
        src.update(other.src)

        # add up stats field
        stats = dict(self.stats)
        for key, value in other.stats.items():
            if key in stats:
                if isinstance(stats[key], (int, float)) and isinstance(value, (int, float)):
                    stats[key] += value
            else:  # new key
                stats[key] = value

        return BiothingHubMeta(
            biothing_type=biothing_type,
            build_date=build_date.isoformat() if isinstance(build_date, datetime) else build_date,
            build_version=build_version,
            src=src,
            stats=stats,
        )



from biothings.web.handlers import BaseAPIHandler, BaseHandler
from biothings.web.options.openapi import OpenAPIDocumentBuilder


class StatusHandler(BaseHandler):
    """Web service health check"""

    # if calling set_status instead of raising exceptions
    # when failure happens, no error will be propogated
    # to sentry monitoring. choose the desired one basing
    # on overall health check architecture.

    # It is technically better to return text or HTML
    # instead of JSON as this is NOT an API endpoint.
    # Instead, it's for automated status check, in which
    # case a simple text response of "OK" suffice, or for
    # human consumption when "dev" parameter is provided,
    # in which case an HTML page is more readible.

    def head(self):
        return self._check()

    async def get(self):
        dev = self.get_argument("dev", None)
        res = await self._check(dev is not None)
        self.finish(res)

    async def _check(self, dev=False):
        try:  # some db connections support async operations
            response = await self.biothings.health.async_check(dev)
        except (AttributeError, NotImplementedError):
            response = self.biothings.health.check()
        return response


class FrontPageHandler(BaseHandler):
    def get(self):
        self.render(
            template_name="home.html",
            alert="Front Page Not Configured.",
            title="Biothings API",
            contents=self.biothings.handlers.keys(),
            support=self.biothings.metadata.types,
            url="http://biothings.io/",
        )

    def get_template_path(self):
        import biothings.web.templates

        return next(iter(biothings.web.templates.__path__))


class APISpecificationHandler(BaseAPIHandler):
    # Proof of concept
    # Not documented for public access

    # There are multiple **correctness** issues
    # For internal use only. Use with caution.

    @staticmethod
    def _type_to_schema(_type):  # for query strings
        _mapping = {
            "list": "array",
            "bool": "boolean",
            "int": "integer",
            "str": "string",
            "float": "number",
        }
        _type = _mapping.get(_type, "object")
        if _type == "array":
            return {
                "type": "array",
                "items": {"type": "string"},
            }
        else:
            return {"type": _type}

    @staticmethod
    def _binds(context, param, option):
        # https://swagger.io/specification/#parameter-object
        location = option.get("location", ("query",))
        _type = option.get("type", str).__name__

        if "query" in location:
            _param = context.parameter(
                param,
                "query",
                option.get("required", False),
            )
            _schema = APISpecificationHandler._type_to_schema(_type)
            if option.get("default"):
                _schema["default"] = option["default"]
            _param.schema(_schema)

    def get(self):
        openapi = OpenAPIDocumentBuilder()
        openapi.info(title="Biothings API", version="0.0.0")
        for path, handler in self.biothings.handlers.items():
            if not issubclass(handler, BaseAPIHandler):
                continue
            if path != "/":
                path = path.rstrip("/?")
            PATH_PARAM = r"(?:/([^/]+))"
            PATH_TOKEN = r"/{id}"
            if PATH_PARAM in path:
                # this is pretty much hard-coded.
                # the corresponding path param is
                # also likely not handled correctly.
                path = path.replace(PATH_PARAM, PATH_TOKEN)
            _path = openapi.path(path)
            optionset = self.biothings.optionsets[handler.name]
            for param, option in optionset.get("*", {}).items():
                self._binds(_path, param, option)
            for method in ("get", "post", "put", "delete"):
                if getattr(handler, method) is type(self)._unimplemented_method:
                    continue
                _method = getattr(_path, method)()
                for param, option in optionset.get(method.upper(), {}).items():
                    self._binds(_method, param, option)
                if PATH_TOKEN in path:
                    for param in _method.document["parameters"]:
                        if param["name"] == "id":  # hard-coded...
                            param["in"] = "path"
                if method == "post":
                    # might be simplified by using "$ref" syntax
                    _method.document["requestBody"] = {
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        key: self._type_to_schema(val.get("type", str).__name__)
                                        for key, val in optionset.get(method.upper(), {}).items()
                                    },
                                }
                            },
                            "application/yaml": {},
                            "application/x-www-form-urlencoded": {},
                            "multipart/form-data": {},
                        }
                    }
        self.finish(openapi.document)

        # internal parameter parsing data structure
        # self.finish(self.biothings.optionsets.log())



""" Some utility functions that do document traversal """
from biothings.utils.common import is_seq

# doc labelled in breadth first way with letters
test_doc = {
    "a": {
        "b": {"e": {"i": "i value", "j": "j value"}, "f": "f value"},
        "c": "c value",
        "d": {"g": {"k": "k value", "l": "l value"}, "h": "h value"},
    }
}


class QueueEmptyError(Exception):
    pass


class StackEmptyError(Exception):
    pass


class Stack(object):
    def __init__(self):
        self.stack = []

    def push(self, obj):
        """put obj on stack"""
        self.stack.insert(0, obj)

    def pop(self):
        try:
            return self.stack.pop(0)
        except IndexError:
            raise StackEmptyError("Can't pop object from an empty stack")

    def isempty(self):
        return len(self.stack) == 0


class Queue(object):
    def __init__(self):
        self.queue = []

    def push(self, obj):
        """put obj on queue"""
        self.queue.append(obj)

    def pop(self):
        """get next obj from queue"""
        try:
            return self.queue.pop(0)
        except IndexError:
            # empty queue
            raise QueueEmptyError("Can't pop object from an empty queue")

    def isempty(self):
        return len(self.queue) == 0


def breadth_first_traversal(doc):
    """Yield a 2 element tuple for every k, v pair in document items (nodes are visited in breadth first order
    k is itself a tuple of keys annotating the path for this node (v) to root
    v is the node value
    """
    return _generic_traversal(doc, Queue)


def depth_first_traversal(doc):
    """Yield a 2 element tuple for every k, v pair in document items (nodes are visited in depth first order
    k is itself a tuple of keys annotating the path for this node (v) to root
    v is the node value
    """
    return _generic_traversal(doc, Stack)


def _generic_traversal(doc, structure):
    _struct = structure()

    # push first level
    for k, v in doc.items():
        # _struct.push((tuple([k]), v))   # TODO: remove this line
        _struct.push(((k,), v))

    while not _struct.isempty():
        _next = _struct.pop()
        yield _next
        if isinstance(_next[1], dict):
            # push this level
            for k, v in _next[1].items():
                _struct.push((tuple(list(_next[0]) + [k]), v))
        elif is_seq(_next[1]):
            # push all elements in a list/tuple
            for o in _next[1]:
                _struct.push((_next[0], o))


def breadth_first_recursive_traversal(doc, path=None):
    """doesn't exactly implement breadth first ordering it seems, not sure why..."""
    # TODO fix this...
    path = path or []
    if isinstance(doc, dict):
        for k, v in doc.items():
            yield (tuple(list(path) + [k]), v)
        for k, v in doc.items():
            yield from breadth_first_recursive_traversal(v, tuple(list(path) + [k]))
    elif is_seq(doc):
        for o in doc:
            yield (tuple(list(path)), o)
        for o in doc:
            yield from breadth_first_recursive_traversal(o, tuple(list(path)))


def depth_first_recursive_traversal(doc, path=None):
    path = path or []
    if isinstance(doc, dict):
        for k, v in doc.items():
            _path = tuple(list(path) + [k])
            yield (_path, v)
            yield from depth_first_recursive_traversal(v, _path)
    elif is_seq(doc):
        for o in doc:
            _path = tuple(list(path))
            yield (_path, o)
            yield from depth_first_recursive_traversal(o, _path)



import inspect
import json
import os
import re
from collections import UserList, UserString, deque
from copy import deepcopy
from dataclasses import asdict, dataclass, field
from importlib import import_module
from json import JSONDecodeError

from biothings.utils.dataload import dict_traverse
from biothings.utils.jsondiff import make as jsondiff
from biothings.utils.loggers import setup_default_log


class ConfigurationError(Exception):
    pass


class ConfigurationValue:
    """
    type to wrap default value when it's code and needs to be interpreted later
    code is passed to eval() in the context of the whole "config" dict
    (so for instance, paths declared before in the configuration file can be used
    in the code passed to eval)
    code will also be executed through exec() *if* eval() raised a syntax error. This
    would happen when code contains statements, not just expression. In that case,
    a variable should be created in these statements (named the same as the original
    config variable) so the proper value can be through ConfigurationManager.
    """

    def __init__(self, code):
        self.code = code

    def get_value(self, name, conf):
        """
        Return value by eval'ing code in self.code, in the context of given configuration
        dict (namespace), for given config parameter name.
        """
        try:
            return eval(self.code, conf.__dict__)
        except SyntaxError:
            # try exec, maybe it's a statement (not just an expression).
            # in that case, it eeans user really knows what he's doing...
            exec(self.code, conf.__dict__)
            # there must be a variable named the same same, in that dict,
            # coming from code's statements
            return conf.__dict__[name]


class ConfigurationDefault:
    def __init__(self, default, desc):
        self.default = default
        self.desc = desc


# the above dynamic value types must be evaluated at runtime
# and not cached, because they may contain reference to the
# current time, which is commonly used in hub operations.


def is_jsonable(x):
    try:
        json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False


def set_default_folder(data_archive_root, sub_folder):
    """set default sub folder based on data_archive_root"""

    # trim off "datasources" to get the data_root folder, otherwise, use it as is.
    # '/data/mydisease/datasources' --> '/data/mydisease/'
    # "datasources" was reserved for downloaded data sources files
    _data_root = (
        data_archive_root[: -len("datasources")] if data_archive_root.endswith("datasources") else data_archive_root
    )
    return os.path.join(_data_root, sub_folder)


class ConfigurationWrapper:
    """
    Wraps and manages configuration access and edit. A singleton
    instance is available throughout all hub apps using biothings.config
    or biothings.hub.config after calling import biothings.hub.
    In addition to providing config value access, either from config files
    or database, config manager can supersede attributes of a class with values
    coming from the database, allowing dynamic configuration of hub's elements.
    """

    # This class contains some of Sebastien's original design
    # It has been moved from biothings.__init__ module to here.

    def __init__(self, default_config, conf):
        """
        When constructing a ConfigurationWrapper instance,
        variables will be defined with default values coming from default_config,
        then they can be overridden by conf's values,
        or new variables will be added if not defined in default_conf.
        Only metadata come from default_config will be used.
        """

        # Store current pid to use later for checking with the runtime's pid.
        self._pid = os.getpid()

        self._module = conf  # python module, typically config.py
        # update self._module with default value for missing configrations
        for attr in dir(default_config):
            default_value = getattr(default_config, attr)
            if not hasattr(self._module, attr):
                setattr(self._module, attr, default_value)

        self._annotations = _parse_comments(default_config, conf)  # section, visibility..
        self._db = None  # typically set by _config_for_app()

        self._modified = False
        self._readonly = True

        if hasattr(self._module, "CONFIG_READONLY"):
            self._readonly = self._module.CONFIG_READONLY

        # When config is readonly, or for readonly config keys,
        # we do a one-time evaluation for those config values set to a ConfigurationDefault instance.
        # This will save repeated code evaluation.
        for attr in dir(self._module):
            if attr == "logger":
                continue  # config.logger will be handled specifically later
            if isinstance(getattr(self._module, attr), ConfigurationDefault):
                if self._readonly or self._annotations.get(attr, {}).get("readonly", False):
                    setattr(self._module, attr, self.get_value_from_file(attr))

        # setup config.logger if not set yet
        logger = getattr(self._module, "logger", None)
        if not logger or isinstance(logger, ConfigurationDefault):
            if hasattr(self._module, "LOG_FOLDER"):
                # set logger if LOG_FOLDER is set
                self._module.logger = setup_default_log(
                    getattr(self._module, "LOGGER_NAME", "hub"),
                    self._module.LOG_FOLDER,
                )
            elif isinstance(logger, ConfigurationDefault):
                # set logger based on default value from default_config
                self._module.logger = self.get_value_from_file("logger")
                self._module.logger.warning('Missing "LOG_FOLDER" setting, default logger set to console only')
            else:
                raise ConfigurationError('Cannot set "config.logger", check "LOG_FOLDER" or "logger" setting')

    @property
    def modified(self):
        if self._db:
            self._modified = bool(self._db.count())
        return self._modified

    @property
    def readonly(self):
        return self._readonly

    def __getattr__(self, name):
        try:
            val = self.get_value_from_db(name)
        except (KeyError, ValueError, AttributeError):
            val = self.get_value_from_file(name)
        return val

    def __delattr__(self, name):
        # TODO I don't think this is a good idea
        delattr(self._module, name)

    def __getitem__(self, name):
        # for dotfield notation,
        # like MyClass.CLS_ATTR
        return self.__getattr__(name)

    def show(self):
        # correspond to /config endpoint
        # the result strcture is designed to
        # maintain compatibility with the frontend

        _config = {}
        _class = {}

        # basic information
        for key in self._annotations:
            x = self._annotations[key]
            # y = serializable(x)

            # "invisible" describes the visibility of keys
            # "hidden" describes the visibility of values

            if x["invisible"]:
                continue

            y = dict(x)
            y["confmod"] = str(y.pop("confmod", ""))
            y["desc"] = y.pop("description")

            default = getattr(self._module, key)
            if is_jsonable(default):
                y["default"] = default

            value = getattr(self, key)
            if is_jsonable(value):
                y["value"] = value

            if self._db.find_one({"_id": key}):
                y["dynamic"] = True

            if x["hidden"]:
                y["value"] = "********"
                y["default"] = "********"

            if y.get("default") and y.get("value"):
                _diff = jsondiff(y["default"], y["value"])
                if _diff:  # frontend shows "reset" button
                    y["diff"] = _diff

            _config[key] = y

        # process-level transient parameters
        for key in _list_attrs(self):
            _config[key] = {
                "value": getattr(self, key),
                "readonly": True,  # wrt to outside
            }

        # transient parameters on self._module are
        # not shown here, like self._module.logger

        # class attr superseding
        for doc in self._db.find():
            key = doc["_id"]
            if len(key.split(".")) == 2:
                _class[key] = json.loads(doc["json"])

        return {
            "scope": {
                "config": _config,
                "class": _class,
            },
            "_dirty": self.modified,
            "allow_edits": not self.readonly,
        }

    def reset(self, name=None):
        if not name:  # global reset
            self._modified = False
            return self._db.remove({})

        res = self._db.remove({"_id": name})
        # res = {'n': 1, 'ok': 1.0}
        return res["ok"]

    def store_value_to_db(self, name, value):
        if not self._db:
            raise RuntimeError("Transient parameter requires DB setup.")
        if self._readonly:
            raise RuntimeError("Configuration is globally read-only.")
        if self._annotations.get(name, {}).get("readonly"):
            raise RuntimeError("Parameter read-only.")
        if self._annotations.get(name, {}).get("invisible"):
            raise RuntimeError("Parameter reserved.")
        if name == "CONFIG_READONLY":  # False -> True also not allowed.
            raise RuntimeError("Runtime modification not allowed.")

        try:  # already json?
            json.loads(value)
        except JSONDecodeError:
            value = json.dumps(value)

        res = self._db.update_one(
            {"_id": name},
            {"$set": {"json": value}},
            upsert=True,
        )

        self._modified = True
        try:  # serializable result for mongo
            return res.raw_result
        except AttributeError:  # for sqlite, etc..
            return res

    def get_value_from_db(self, name):
        if not self._db:  # without db, only support module params.
            raise AttributeError("Transient parameter requires DB setup.")

        # Our db instance is shared with child process when this process is folked, and that action it not fork-safe.
        # So we must recreate db instance in the child process.
        # Ref: https://pymongo.readthedocs.io/en/stable/faq.html#id3
        if os.getpid() != self._pid:
            self._pid = os.getpid()
            self._db = None
            self._db = self._get_db_function()

        doc = self._db.find_one({"_id": name})
        if not doc:
            raise AttributeError(name)

        val = json.loads(doc["json"])
        return val

    def get_value_from_file(self, name):
        # raw value might require eval
        val = getattr(self._module, name)

        def eval_default_value(k, v):
            if isinstance(v, ConfigurationDefault):
                if isinstance(v.default, ConfigurationValue):
                    return (k, v.default.get_value(name, self._module))
                else:
                    return (k, v.default)
            elif isinstance(v, ConfigurationValue):
                return (k, v.get_value(k, self._module))
            elif isinstance(v, ConfigurationError):
                # default ConfigurationError value needs to be
                # set in hub's config file
                raise v
            else:
                return (k, v)

        if isinstance(val, dict):
            # walk the dict and instantiate values when special
            dict_traverse(val, eval_default_value, traverse_list=True)
        else:
            # just use the same func but ignore "k" key, not a dict
            # pass unhashable "k" to make sure we'd raise an error
            # while dict traversing  if we're not supposed to be here
            _, val = eval_default_value({}, val)

        return val

    def supersede(self, klass):
        """supersede class variable with db values"""

        if not isinstance(klass, type):
            raise TypeError("Don't know how to supersede type '%s'" % type(klass))

        for doc in self._db.find():
            attr = doc["_id"]  # MyClass.EXAMPLE_CLASS_VAR
            if attr.startswith(klass.__name__):
                splits = attr.split(".")
                if len(splits) == 2:
                    value = json.loads(doc["json"])
                    setattr(klass, splits[1], value)

    def __repr__(self):
        return "<%s over %s>" % (self.__class__.__name__, str(self._module))


def _list_attrs(conf_mod):
    attrs = set()  # hub-supported attributes
    for attrname in dir(conf_mod):
        # re pattern to find config param
        # (by convention, all upper caps, _ allowed, that's all)
        if re.compile("^([A-Z_]+)$").match(attrname):
            attrs.add(attrname)
    return attrs


def _parse_comments(default_conf_mod, conf_mod):
    """
    TODO NEED REVIEW

    Parse configuration module and extract documentation from default_conf_mod.
    Documentation can be found in different place (in order):
    1. the configuration value is a ConfigurationDefault instance (specify a default value)
       or a ConfigurationError instance, in which case the documentation is taken
       from the instance doc.
    2. the documentation can be specified as an inline comment
    3. the documentation can be specified as comments above

    If the configuration module also import another (or more) config modules, those
    modules will be searched as well, if nothing could be found in the main module.
    As soon as a documentation is found, the search stops (importance of module imports order)

    There are several special comment formats used to organize and manager params:
    - all comments above a param are considered as documentation/description for the parameter,
      until a empty line or a non-comment line is found.
    - to be considered as documentation, comments also need to have a space after "#". So:
            # my comment
      will be kepts as documentation, but:
            #my comment
      will just be ignored (so python code can be commented that way, without then being part
      of the documentation)
    - A section title can be added to organize parameters. A section title looks like this:
            #* my section title *#
      It can be added for each parameters, or the first section found above is the section
      the parameter will be associated to. An empty section title can be used to reset the
      section title and associate the current parameter to default one:
            #* *#
      If no section is found, all parameters are part of the default one (None).
    - some parameters needs to be kept secret (like passwords for instance):
            #- invisible -#
      will hide the parameter, including the name, value, description, from the configuration
            #- hidden -#
      will keep the parameter in the configuration displayed to users, but its value will be omitted
            #- readonly -#
      will allow the parameter to shown, but not editable

      Note: special comments can stacked:
            #- readonly -#
            #- hidden -#
      will make the parameter read-only, and its value won't be displayed
    """
    attrs = _list_attrs(default_conf_mod)
    attrs.update(_list_attrs(conf_mod))

    try:
        configs = []
        _configs = deque([default_conf_mod, conf_mod])

        while _configs:
            config = _configs.popleft()
            lines = inspect.getsourcelines(config)[0]

            # ignore special comment lines which donot come from default_config_mod
            if config != default_conf_mod:
                lines = [line for line in lines if not line.strip().startswith("#")]

            lines = ConfigLines(ConfigLine(line) for line in lines)

            pat = re.compile(r"^from\s+(.*?)\s+import\s+\*")
            for line in lines:
                match = pat.match(line.data)
                if match:
                    base = match.groups()[0]
                    base_mod = import_module(base)
                    _configs.append(base_mod)

            result = lines.parse(attrs)
            configs.append((config, result))

        _info = {}  # merged
        for config, result in reversed(configs):
            for attr, meta in result.items():
                meta.confmod.feed(config)
                if attr in _info:
                    _info[attr].update(meta)
                else:
                    _info[attr] = meta

        return {k: v.asdict() for k, v in _info.items()}
    except Exception:
        return dict.fromkeys(attrs, {})


class MetaField:
    default = type(None)

    def __init__(self, value=None):
        self._value = value or self.default()

    @property
    def value(self):
        return self._value

    def feed(self, value):
        self._value = value

    def clear(self):
        self._value = self.default()


class Text(MetaField):
    def feed(self, value):
        assert isinstance(value, str)
        self._value = value.strip() or None


class Flag(MetaField):
    default = bool

    def feed(self, value):
        if value:  # cannot unset a flag
            self._value = value


class Paragraph(MetaField):
    default = list

    @property
    def value(self):
        _value = " ".join(self._value)
        return _value or None

    def feed(self, value):
        if isinstance(value, Paragraph):
            value = value.value
        assert isinstance(value, str)
        self._value.append(value.strip())


# TODO currently not processing SKIPs
# I noticed SKIPs in the config file
# But it seems that it wasn't processed
# in the previous version either


@dataclass
class ConfigAttrMeta:
    confmod: MetaField = field(default_factory=MetaField)
    section: Text = field(default_factory=Text)  # persistent
    description: Paragraph = field(default_factory=Paragraph)
    readonly: Flag = field(default_factory=Flag)
    hidden: Flag = field(default_factory=Flag)
    invisible: Flag = field(default_factory=Flag)

    def update(self, meta):
        assert isinstance(meta, ConfigAttrMeta)
        for _field, value in meta.asdict().items():
            self.feed(_field, value)

    def asdict(self):
        confmod, self.confmod = self.confmod, MetaField()
        result = {k: v.value for k, v in asdict(self).items()}
        result["confmod"] = confmod.value  # cannot pickle in asdict
        return result

    # ------------------------------------

    def feed(self, field, value):
        if field and value is not None:
            getattr(self, field).feed(value)

    def reset(self):
        self.description.clear()
        self.readonly.clear()
        self.hidden.clear()
        self.invisible.clear()

    def commit(self):
        result = deepcopy(self)
        self.reset()
        return result


class ConfigLines(UserList):
    def parse(self, attrs=()):
        result = {}
        attrs = set(attrs)
        current = ConfigAttrMeta()

        for line in self.data:
            # feed
            field, value = line.match()
            current.feed(field, value)

            # commit
            if "=" in line:
                attr = line.split("=")[0].strip()
                if attr in attrs:
                    result[attr] = current.commit()

            # reset
            if not line.strip():
                current.reset()

        return result


class ConfigLine(UserString):
    PATTERNS = (
        ("hidden", re.compile(r"^#\s?-\s*hide\s*-\s?#\s*$"), lambda m: bool(m)),
        ("invisible", re.compile(r"^#\s?-\s*invisible\s*-\s?#\s*$"), lambda m: bool(m)),
        ("readonly", re.compile(r"^#\s?-\s*readonly\s*-\s?#\s*$"), lambda m: bool(m)),
        ("section", re.compile(r"^#\s?\*\s*(.*)\s*\*\s?#\s*$"), lambda m: m.groups()[0]),
        ("description", re.compile(r".*\s*#\s+(.*)$"), lambda m: m.groups()[0]),
    )

    def match(self):
        for _type, pattern, func in self.PATTERNS:
            match = pattern.match(self.data)
            if match:
                return _type, func(match)
        return None, None



import asyncio
import datetime
import gzip
import logging
import os
from collections import OrderedDict, UserList
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from functools import partial
from itertools import chain
from logging.handlers import TimedRotatingFileHandler
from threading import Thread
from types import MethodType
from typing import NamedTuple, Union

import requests

LOG_FORMAT_STRING = "%(asctime)s [%(process)d:%(threadName)s] - %(name)s - %(levelname)s -- %(message)s"
DATEFMT = "%H:%M:%S"


class GZipRotator:
    def __call__(self, source, dest):
        os.rename(source, dest)
        f_in = open(dest, "rb")
        f_out = gzip.open("%s.gz" % dest, "wb")
        f_out.writelines(f_in)
        f_out.close()
        f_in.close()
        os.remove(dest)


def create_logger(log_folder, logger_name, level=logging.DEBUG):
    """
    Create and return a file logger if log_folder is provided.
    If log_folder is None, no file handler will be created.
    """
    if log_folder:
        if not os.path.exists(log_folder):
            os.makedirs(log_folder)
        logfile = os.path.join(log_folder, logger_name)
        if not logfile.endswith(".log"):
            logfile += ".log"
    else:
        logfile = None
    logger = logging.getLogger(logger_name)
    logger.setLevel(level)
    return logger, logfile


def configurate_file_handler(logger, logfile, formater=None, force=False):
    if not formater:
        formater = logging.Formatter(LOG_FORMAT_STRING, datefmt=DATEFMT)

    fh = TimedRotatingFileHandler(logfile, when="D")
    fh.setFormatter(formater)
    fh.rotator = GZipRotator()
    fh.name = "logfile"
    if force or fh.name not in [h.name for h in logger.handlers]:
        logger.addHandler(fh)


def setup_default_log(default_logger_name, log_folder, level=logging.DEBUG):
    # this will affect any logging calls
    logging.basicConfig(level=level)
    logger, logfile = create_logger(log_folder, default_logger_name, level=level)
    if logfile:
        configurate_file_handler(logger, logfile)
    return logger


def get_logger(logger_name, log_folder=None, handlers=("console", "file", "slack"), timestamp=None, force=False):
    """
    Configure a logger object from logger_name and return (logger, logfile)
    """
    from biothings import config as btconfig

    if timestamp:
        raise DeprecationWarning("Timestamp is deprecated")

    if not log_folder:
        log_folder = btconfig.LOG_FOLDER
    # this will affect any logging calls
    logger, logfile = create_logger(log_folder, logger_name)
    fmt = logging.Formatter(LOG_FORMAT_STRING, datefmt=DATEFMT)
    if "file" in handlers and logfile:
        configurate_file_handler(logger, logfile, formater=fmt, force=force)

    if "hipchat" in handlers:
        raise DeprecationWarning("Hipchat is dead...")

    if "slack" in handlers and getattr(btconfig, "SLACK_WEBHOOK", None):
        nh = SlackHandler(
            btconfig.SLACK_WEBHOOK,
            getattr(btconfig, "SLACK_MENTIONS", []),
        )
        nh.setFormatter(fmt)
        nh.name = "slack"
        if force or nh.name not in [h.name for h in logger.handlers]:
            logger.addHandler(nh)

    def notify(self, *args, **kwargs):
        return self.info(*args, **kwargs, extra={"notify": True})

    logger.notify = MethodType(notify, logger)

    return (logger, logfile)


class EventRecorder(logging.StreamHandler):
    def __init__(self, *args, **kwargs):
        super(EventRecorder, self).__init__(*args, **kwargs)
        from biothings.utils.hub_db import get_event

        self.eventcol = get_event()

    def emit(self, record):
        async def aioemit(msg):
            def recorded(f):
                # res = f.result()
                f.result()

            fut = loop.run_in_executor(None, partial(self.eventcol.save, msg))
            fut.add_done_callback(recorded)
            await fut
            return fut

        if record.__dict__.get("notify") or record.__dict__.get("event"):
            try:
                loop = asyncio.get_event_loop()
                msg = {
                    "_id": record.created,
                    "asctime": record.asctime,
                    "msg": record.message,
                    "level": record.levelname,
                    "name": record.name,
                    "pid": record.process,
                    "pname": record.processName,
                }
                fut = aioemit(msg)
                asyncio.ensure_future(fut)
            except Exception as e:
                logging.error("Couldn't record event: %s", e)


class WSLogHandler(logging.StreamHandler):
    """
    when listener is a bt.hub.api.handlers.ws.LogListener instance,
    log statements are propagated through existing websocket
    """

    def __init__(self, listener):
        super().__init__()
        self.listener = listener
        self.count = 0

    def payload(self, record):
        msg = self.format(record)
        return {
            "_id": self.count,
            "op": "log",
            "msg": msg,
            "logger": record.name,
            "level": record.levelname,
            "ts": datetime.datetime.now().isoformat(),
        }

    def emit(self, record):
        self.count += 1
        self.listener.read(self.payload(record))


class WSShellHandler(WSLogHandler):
    """
    when listener is a bt.hub.api.handlers.ws.LogListener instance,
    log statements are propagated through existing websocket
    """

    def payload(self, record):
        types = {ShellLogger.INPUT: "input", ShellLogger.OUTPUT: "output"}
        return {
            "_id": self.count,
            "op": "shell",
            "cmd": record.msg,
            "type": types.get(record.levelno, "unknown"),
            "ts": datetime.datetime.now().isoformat(),
        }


class ShellLogger(logging.Logger):
    """
    Custom "levels" for input going to the shell
    and output coming from it (just for naming)
    """

    OUTPUT = 1000
    INPUT = 1001

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.manager.loggerDict[self.name] = self

    def input(self, msg, *args, **kwargs):
        self._log(self.__class__.INPUT, msg, args, **kwargs)

    def output(self, msg, *args, **kwargs):
        self._log(self.__class__.OUTPUT, msg, args, **kwargs)


# -----------------------------------------------
#                SLACK INTEGRATION
# -----------------------------------------------
# https://api.slack.com/reference/block-kit/blocks
# https://api.slack.com/reference/block-kit/composition-objects#text
# https://api.slack.com/reference/messaging/attachments


class Squares(Enum):
    CRITICAL = ":large_purple_square:"
    ERROR = ":large_red_square:"
    WARNING = ":large_orange_square:"
    INFO = ":large_blue_square:"
    DEBUG = ":white_large_square:"
    NOTSET = ""


class Colors(Enum):
    CRITICAL = "#7b0099"
    ERROR = "danger"  # red
    WARNING = "warning"  # yellow
    INFO = "good"  # green
    DEBUG = "#a1a1a1"
    NOTSET = "#d6d2d2"


@dataclass
class Range:
    start: Union[int, float] = 0  # inclusive
    end: Union[int, float] = float("inf")  # exclusive


class Record(NamedTuple):
    range: Range
    value: Enum


class LookUpList(UserList):
    def __init__(self, initlist):
        super().__init__(initlist)
        assert all(isinstance(x, Record) for x in self.data)
        assert all(isinstance(x.range, Range) for x in self.data)

    def find_index(self, val):
        _l, _r = 0, len(self.data)
        while _l < _r:
            mid = (_l + _r) // 2
            start = self.data[mid].range.start
            end = self.data[mid].range.end
            if val < start:
                _r = mid
            elif val >= end:
                _l = mid + 1
            else:  # found
                return mid

    def find(self, val):
        index = self.find_index(val)
        if index is not None:
            return self.data[index].value


ICONS = LookUpList(
    [
        Record(Range(float("-inf"), logging.DEBUG), Squares.NOTSET),
        Record(Range(logging.DEBUG, logging.INFO), Squares.DEBUG),
        Record(Range(logging.INFO, logging.WARNING), Squares.INFO),
        Record(Range(logging.WARNING, logging.ERROR), Squares.WARNING),
        Record(Range(logging.ERROR, logging.CRITICAL), Squares.ERROR),
        Record(Range(logging.CRITICAL, float("inf")), Squares.CRITICAL),
    ]
)

COLORS = LookUpList(
    [
        Record(Range(float("-inf"), logging.DEBUG), Colors.NOTSET),
        Record(Range(logging.DEBUG, logging.INFO), Colors.DEBUG),
        Record(Range(logging.INFO, logging.WARNING), Colors.INFO),
        Record(Range(logging.WARNING, logging.ERROR), Colors.WARNING),
        Record(Range(logging.ERROR, logging.CRITICAL), Colors.ERROR),
        Record(Range(logging.CRITICAL, float("inf")), Colors.CRITICAL),
    ]
)


class SlackMessage:
    def __init__(self):
        self._blocks = []
        self._attachments = []

    def markdown(self, text, prefixes=(), suffixes=()):
        text = text.strip()
        text = " ".join((*filter(None, prefixes), text))
        text = " ".join((text, *filter(None, suffixes)))
        if text:
            # empty value causes 400
            # error in slack API
            self._blocks.append({"type": "section", "text": {"type": "mrkdwn", "text": text}})

    def plaintext(self, text, color):
        text = text.strip()
        if text:
            # empty value causes 400
            # error in slack API
            self._attachments.append({"text": text, "color": color})

    def build(self):
        return deepcopy({"blocks": self._blocks, "attachments": self._attachments})


class ReceiverGroup(UserList):
    def __init__(self, initlist=None):
        super().__init__(initlist or [])
        self.prev = None
        self.range = Range()


class SlackMentionPolicy:
    # TODO Support string representation
    # of a level throughout this class

    def __init__(self, policy):
        if isinstance(policy, dict):
            assert all(isinstance(lvl, int) for lvl in policy.keys())
            assert all(isinstance(m, (list, tuple)) for m in policy.values())
            assert all(isinstance(p, str) for p in chain(*policy.values()))

            # TODO maybe a user should only subscribe to
            # one logging level considering how logging
            # propagation works?

            self._policy = OrderedDict(sorted((level, ReceiverGroup(receivers)) for level, receivers in policy.items()))

        elif isinstance(policy, (tuple, list)):
            assert all(isinstance(m, str) for m in policy)
            assert len(set(policy)) == len(policy)

            self._policy = OrderedDict({logging.ERROR: ReceiverGroup(policy)})

        elif isinstance(policy, str):
            self._policy = OrderedDict({logging.ERROR: ReceiverGroup([policy])})

        else:  # see test cases for supported values.
            raise TypeError("Unsupported Slack Mentions.")

        _prev = None
        for level, mention in self._policy.items():
            mention.prev = _prev
            mention.range.start = level
            if _prev:
                # previously configured logging receivers handle
                # all levels until the next highest log level
                _prev.range.end = level
            _prev = mention

    def mentions(self, level):
        assert isinstance(level, int)

        if level not in self._policy:
            # a logging level can be an integer between
            # the commonly defined, named ones, search for
            # which range it belongs if that's the case
            levels = list(self._policy.values())
            _l, _r = 0, len(levels)
            while _l < _r:
                mid = (_l + _r) // 2
                start = levels[mid].range.start
                end = levels[mid].range.end
                if level < start:
                    _r = mid
                elif level >= end:
                    _l = mid + 1
                else:  # found
                    level = start
                    break

        # all receivers that should be
        # mentioned for this log level
        mentions = []

        # TODO what should we do if we encounter
        # duplicated entries due to bad configurations

        if level in self._policy:
            current = self._policy[level]
            while current:
                mentions.extend(current)
                current = current.prev

        return mentions


class SlackHandler(logging.StreamHandler):
    def __init__(self, webhook, mentions):
        super(SlackHandler, self).__init__()
        self.webhook = webhook
        self.mentions = SlackMentionPolicy(mentions)

    @staticmethod
    def send(webhook, message, level, mentions=()):  # logging entry point
        msg = SlackMessage()
        msg.markdown("", suffixes=mentions)
        msg.plaintext(message, COLORS.find(level).value)
        res = requests.post(webhook, json=msg.build())
        res.raise_for_status()

    def emit(self, record):
        if record.__dict__.get("notify"):
            Thread(
                target=SlackHandler.send,  # blocking
                args=(
                    self.webhook,
                    self.format(record),
                    record.levelno,
                    self.mentions.mentions(record.levelno),
                ),
            ).start()



import datetime
import os
from dataclasses import dataclass, field

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID


def generate_ssh_key(path):
    ssh_privkey_filename = path
    ssh_pubkey_filename = f"{path}.pub"

    print("Generating SSH Key ...")
    privkey = rsa.generate_private_key(65537, 2048)
    with open(ssh_privkey_filename, "wb") as f:
        f.write(
            privkey.private_bytes(
                serialization.Encoding.PEM,
                serialization.PrivateFormat.OpenSSH,
                serialization.NoEncryption(),
            )
        )
    pubkey = privkey.public_key().public_bytes(serialization.Encoding.OpenSSH, serialization.PublicFormat.OpenSSH)
    with open(ssh_pubkey_filename, "wb") as f:
        f.write(pubkey)
    print("SSH Key has been generated, Public Key:\n")
    print(pubkey.decode("ASCII"))
    print()


@dataclass
class CertificateInformation:
    country: str
    state_or_province: str
    locality: str
    organization: str
    common_name: str

    passphare: str = None
    dns_names: list[str] = field(default_factory=list)


def generate_self_signed_cert_file(dir_path, **cert_info: CertificateInformation):
    """
    This helper method generates SSL certificate, which can be passed to TORNADO_SETTINGS to support HTTPS.
    The pkcs12 file should be added to browser, but we still need to manually tell browser to allow us access our hub.

    Example Tornado settings:
    ```
        import ssl

        ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_ctx.load_cert_chain(
            os.path.join(path_to_cert_file),
            os.path.join(path_to_key_file),
        )

        TORNADO_SETTINGS = {
            "ssl_options": ssl_ctx
        }
    ```

    Ref:
    https://cryptography.io/en/latest/x509/tutorial/#creating-a-self-signed-certificate
    https://cryptography.io/en/latest/hazmat/primitives/asymmetric/serialization/#cryptography.hazmat.primitives.serialization.pkcs12.serialize_key_and_certificates
    """  # NOQA

    key_filename = os.path.join(dir_path, "key.pem")
    cert_filename = os.path.join(dir_path, "cert.pem")
    pkcs12_filename = os.path.join(dir_path, "cert.p12")
    cert_info = CertificateInformation(**cert_info)

    print("Generating private key ...")
    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )

    # Write our key to disk for safe keeping
    print("Write private key to file ...")
    encryption_algorithm = serialization.NoEncryption()
    if cert_info.passphare:
        encryption_algorithm = serialization.BestAvailableEncryption(cert_info.passphare.encode())
    with open(key_filename, "wb") as f:
        f.write(
            key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=encryption_algorithm,
            )
        )

    # Various details about who we are. For a self-signed certificate the
    # subject and issuer are always the same.
    print("Generating certificate ...")
    subject = issuer = x509.Name(
        [
            x509.NameAttribute(NameOID.COUNTRY_NAME, cert_info.country),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, cert_info.state_or_province),
            x509.NameAttribute(NameOID.LOCALITY_NAME, cert_info.locality),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, cert_info.organization),
            x509.NameAttribute(NameOID.COMMON_NAME, cert_info.common_name),
        ]
    )
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(
            # Our certificate will be valid for 10 days
            datetime.datetime.utcnow()
            + datetime.timedelta(days=10)
        )
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName(dns_name) for dns_name in cert_info.dns_names or ["localhost"]]),
            critical=False,
            # Sign our certificate with our private key
        )
        .sign(key, hashes.SHA256())
    )

    # Write our certificate out to disk.
    print("Writing certificate to file ...")
    with open(cert_filename, "wb") as f:
        f.write(cert.public_bytes(serialization.Encoding.PEM))

    # Generate PKCS12 to import to browser.
    print("Generate PKCS12 ...")
    pkcs12 = serialization.pkcs12.serialize_key_and_certificates(b"", key, cert, None, serialization.NoEncryption())

    # Write PKCS12 file to import to browser.
    print("Writing PKCS12 to file ...")
    with open(pkcs12_filename, "wb") as f:
        f.write(pkcs12)

    print("Done.")



import asyncio
import json
import os
import time
from collections import UserDict, UserString
from dataclasses import dataclass
from datetime import datetime
from functools import partial

import boto3
from config import logger as logging
from elasticsearch import Elasticsearch
from elasticsearch.exceptions import TransportError

from biothings import config as btconfig
from biothings.hub import SNAPSHOOTER_CATEGORY
from biothings.hub.databuild.buildconfig import AutoBuildConfig
from biothings.hub.datarelease import set_pending_to_release_note
from biothings.utils.common import merge
from biothings.utils.exceptions import RepositoryVerificationFailed
from biothings.utils.hub import template_out
from biothings.utils.hub_db import get_src_build
from biothings.utils.loggers import get_logger
from biothings.utils.manager import BaseManager

from . import snapshot_cleanup as cleaner, snapshot_registrar as registrar
from .snapshot_repo import Repository
from .snapshot_task import Snapshot


class ProcessInfo:
    """
    JobManager Process Info.
    Reported in Biothings Studio.
    """

    def __init__(self, env):
        self.env_name = env

    def get_predicates(self):
        return []

    def get_pinfo(self, step, snapshot, description=""):
        pinfo = {
            "__predicates__": self.get_predicates(),
            "category": SNAPSHOOTER_CATEGORY,
            "step": f"{step}:{snapshot}",
            "description": description,
            "source": self.env_name,
        }
        return pinfo


@dataclass
class CloudStorage:
    type: str
    access_key: str
    secret_key: str
    region: str = "us-west-2"

    def get(self):
        if self.type == "aws":
            session = boto3.Session(
                aws_access_key_id=self.access_key,
                aws_secret_access_key=self.secret_key,
                region_name=self.region,
            )
            return session.resource("s3")  # [X]
        raise ValueError(self.type)


class Bucket:
    def __init__(self, client, bucket, region=None):
        self.client = client  # boto3.S3.Client [X]
        self.bucket = bucket  # bucket name
        self.region = region

    def exists(self):
        bucket = self.client.Bucket(self.bucket)
        return bool(bucket.creation_date)

    def create(self, acl="private"):
        # https://boto3.amazonaws.com/v1/documentation/api
        # /latest/reference/services/s3.html
        # #S3.Client.create_bucket

        return self.client.create_bucket(
            ACL=acl,
            Bucket=self.bucket,
            CreateBucketConfiguration={"LocationConstraint": self.region},
        )

    def __str__(self):
        return f"<Bucket {'READY' if self.exists() else 'MISSING'} name='{self.bucket}' client={self.client}>"


class _UserString(UserString):
    def __str__(self):
        return f"{type(self).__name__}({self.data})"


class TemplateStr(_UserString):
    ...


class RenderedStr(_UserString):
    ...


class RepositoryConfig(UserDict):
    """
    {
        "type": "s3",
        "name": "s3-$(Y)",
        "settings": {
            "bucket": "<SNAPSHOT_BUCKET_NAME>",
            "base_path": "mynews.info/$(Y)",  # per year
        }
    }
    """

    @property
    def repo(self):
        return self["name"]

    @property
    def bucket(self):
        return self["settings"]["bucket"]

    @property
    def region(self):
        return self["settings"]["region"]

    def format(self, doc=None):
        """Template special values in this config.

        For example:
        {
            "bucket": "backup-$(Y)",
            "base_path" : "snapshots/%(_meta.build_version)s"
        }
        where "_meta.build_version" value is taken from doc in
        dot field notation, and the current year replaces "$(Y)".
        """
        template = TemplateStr(json.dumps(self.data))
        string = RenderedStr(template_out(template.data, doc or {}))

        if "%" in string:
            logging.error(template)
            logging.error(string)
            raise ValueError("Failed to template.")

        if template != string:
            logging.debug(template)
            logging.debug(string)

        return RepositoryConfig(json.loads(string.data))


class _SnapshotResult(UserDict):
    def __str__(self):
        return f"{type(self).__name__}({str(self.data)})"


class CumulativeResult(_SnapshotResult):
    ...


class StepResult(_SnapshotResult):
    ...


class SnapshotEnv:
    def __init__(self, job_manager, cloud, repository, indexer, **kwargs):
        self.job_manager = job_manager

        self.cloud = CloudStorage(**cloud).get()
        self.repcfg = RepositoryConfig(repository)
        self.client = Elasticsearch(**indexer["args"])

        self.name = kwargs["name"]  # snapshot env
        self.idxenv = indexer["name"]  # indexer env

        self.pinfo = ProcessInfo(self.name)
        self.wtime = kwargs.get("monitor_delay", 15)

    def _doc(self, index):
        doc = get_src_build().find_one({f"index.{index}.environment": self.idxenv})
        if not doc:  # not asso. with a build
            raise ValueError("Not a hub-managed index.")
        return doc  # TODO UNIQUENESS

    def setup_log(self, index):
        build_doc = self._doc(index)
        log_name = build_doc["target_name"] or build_doc["_id"]
        log_folder = os.path.join(btconfig.LOG_FOLDER, "build", log_name, "snapshot") if btconfig.LOG_FOLDER else None
        self.logger, _ = get_logger(index, log_folder=log_folder, force=True)

    def snapshot(self, index, snapshot=None, recreate_repo=False):
        self.setup_log(index)

        async def _snapshot(snapshot):
            x = CumulativeResult()
            build_doc = self._doc(index)
            cfg = self.repcfg.format(build_doc)
            for step in ("pre", "snapshot", "post"):
                state = registrar.dispatch(step)  # _TaskState Class
                state = state(get_src_build(), build_doc.get("_id"))
                self.logger.info(state)
                state.started()

                job = await self.job_manager.defer_to_thread(
                    self.pinfo.get_pinfo(step, snapshot),
                    partial(getattr(self, state.func), cfg, index, snapshot, recreate_repo=recreate_repo),
                )
                try:
                    dx = await job
                    dx = StepResult(dx)
                except RepositoryVerificationFailed as ex:
                    self.logger.exception(ex)
                    state.failed(snapshot, detail=ex.args)
                    raise ex
                except Exception as exc:
                    self.logger.exception(exc)
                    state.failed({}, exc)
                    raise exc
                else:
                    merge(x.data, dx.data)
                    self.logger.info(dx)
                    self.logger.info(x)
                    state.succeed({snapshot: x.data}, res=dx.data)
            return x

        future = asyncio.ensure_future(_snapshot(snapshot or index))
        future.add_done_callback(self.logger.debug)
        return future

    def pre_snapshot(self, cfg, index, snapshot, **kwargs):
        bucket = Bucket(self.cloud, cfg.bucket, region=cfg.region)
        repo = Repository(self.client, cfg.repo)

        self.logger.info(bucket)
        self.logger.info(repo)

        if kwargs.get("recreate_repo"):
            self.logger.info("Delete old repository")
            repo.delete()

        if not repo.exists():
            if not bucket.exists():
                bucket.create(cfg.get("acl"))
                self.logger.info(bucket)
            repo.create(**cfg)
            self.logger.info(repo)

        try:
            repo.verify(config=cfg)
        except TransportError as tex:
            raise RepositoryVerificationFailed({"error": tex.error, "detail": tex.info["error"]})

        return {
            "__REPLACE__": True,
            "conf": {"repository": cfg.data},
            "indexer_env": self.idxenv,
            "environment": self.name,
        }

    def _snapshot(self, cfg, index, snapshot, **kwargs):
        snapshot = Snapshot(self.client, cfg.repo, snapshot)
        self.logger.info(snapshot)

        _replace = False
        if snapshot.exists():
            snapshot.delete()
            self.logger.info(snapshot)
            _replace = True

        # ------------------ #
        snapshot.create(index)
        # ------------------ #

        while True:
            self.logger.info(snapshot)
            state = snapshot.state()

            if state == "FAILED":
                raise ValueError(state)
            elif state == "IN_PROGRESS":
                time.sleep(self.wtime)
            elif state == "SUCCESS":
                break
            else:  # PARTIAL/MISSING/N/A
                raise ValueError(state)

        return {
            "index_name": index,
            "replaced": _replace,
            "created_at": datetime.now().astimezone(),
        }

    def post_snapshot(self, cfg, index, snapshot, **kwargs):
        build_id = self._doc(index)["_id"]
        set_pending_to_release_note(build_id)
        return {}


class SnapshotManager(BaseManager):
    """
    Hub ES Snapshot Management

    Config Ex:

    # env.<name>:
    {
        "cloud": {
            "type": "aws",  # default, only one supported.
            "access_key": <------------------>,
            "secret_key": <------------------>,
            "region": "us-west-2"
        },
        "repository": {
            "name": "s3-$(Y)",
            "type": "s3",
            "settings": {
                "bucket": "<SNAPSHOT_BUCKET_NAME>",
                "base_path": "mygene.info/$(Y)",  # year
            },
            "acl": "private",
        },
        "indexer": {
            "name": "local",
            "args": {
                "timeout": 100,
                "max_retries": 5
            }
        },
        "monitor_delay": 15,
    }
    """

    def __init__(self, index_manager, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index_manager = index_manager
        self.snapshot_config = {}

    @staticmethod
    def pending_snapshot(build_name):
        src_build = get_src_build()
        src_build.update(
            {"_id": build_name},
            {"$addToSet": {"pending": "snapshot"}},
        )

    # Object Lifecycle Calls
    # --------------------------
    # manager = IndexManager(job_manager)
    # manager.clean_stale_status() # in __init__
    # manager.configure(config)

    def clean_stale_status(self):
        registrar.audit(get_src_build(), logging)

    def configure(self, conf):
        self.snapshot_config = conf
        for name, envdict in conf.get("env", {}).items():
            # Merge Indexer Config
            # ----------------------------------------
            dx = envdict["indexer"]

            if isinstance(dx, str):  # {"indexer": "prod"}
                dx = dict(name=dx)  # .          ↓
            if not isinstance(dx, dict):  # {"indexer": {"name": "prod"}}
                raise TypeError(dx)

            # compatibility with previous hubs.
            dx.setdefault("name", dx.pop("env", None))

            x = self.index_manager[dx["name"]]
            x = dict(x)  # merge into a copy
            merge(x, dx)  # <-

            envdict["indexer"] = x
            # ------------------------------------------
            envdict["name"] = name

            self.register[name] = SnapshotEnv(self.job_manager, **envdict)

    def poll(self, state, func):
        super().poll(state, func, col=get_src_build())

    # Features
    # -----------

    def snapshot(self, snapshot_env, index, snapshot=None, recreate_repo=False):
        """
        Create a snapshot named "snapshot" (or, by default, same name as the index)
        from "index" according to environment definition (repository, etc...) "env".
        """
        env = self.register[snapshot_env]
        return env.snapshot(index, snapshot, recreate_repo=recreate_repo)

    def snapshot_a_build(self, build_doc):
        """
        Create a snapshot basing on the autobuild settings in the build config.
        If the build config associated with this build has:
        {
            "autobuild": {
                "type": "snapshot", // implied when env is set. env must be set.
                "env": "local" // which es env to make the snapshot.
            },
            ...
        }
        Attempt to make a snapshot for this build on the specified es env "local".
        """

        async def _():
            autoconf = AutoBuildConfig(build_doc["build_config"])
            env = autoconf.auto_build.get("env")
            assert env, "Unknown autobuild env."

            if isinstance(env, str):
                indexer_env = env
                snapshot_env = env
            else:  # assume env is an (x,y) pair
                indexer_env, snapshot_env = env

            try:  # find the index (latest) to snapshot
                latest_index = list(build_doc["index"].keys())[-1]

            except Exception:  # no existing indices, need to create one
                await self.index_manager.index(indexer_env, build_doc["_id"])
                latest_index = build_doc["_id"]  # index_name is build name

            return self.snapshot(snapshot_env, latest_index)

        return asyncio.ensure_future(_())

    def snapshot_info(self, env=None, remote=False):
        return self.snapshot_config

    def list_snapshots(self, env=None, **filters):
        return cleaner.find(  # filters support dotfield.
            get_src_build(),
            env=env,
            group_by="build_config",
            return_db_cols=True,
            **filters,
        )

    def cleanup(
        self,
        env=None,  # a snapshot environment describing a repository
        keep=3,  # the number of most recent snapshots to keep in one group
        group_by="build_config",  # the attr of which its values form groups
        dryrun=True,  # display the snapshots to be deleted without deleting them
        **filters,  # a set of criterions to limit which snapshots are to be cleaned
    ):
        """Delete past snapshots and keep only the most recent ones.

        Examples:
            >>> snapshot_cleanup()
            >>> snapshot_cleanup("s3_outbreak")
            >>> snapshot_cleanup("s3_outbreak", keep=0)
        """

        snapshots = cleaner.find(get_src_build(), env, keep, group_by, **filters)  # filters support dotfield.

        if dryrun:
            return "\n".join(
                (
                    "-" * 75,
                    cleaner.plain_text(snapshots),
                    "-" * 75,
                    "DRYRUN ONLY - APPLY THE ACTIONS WITH:",
                    "   > snapshot_cleanup(..., dryrun=False)",
                )
            )

        # return the number of snapshots successfully deleted
        return cleaner.delete(get_src_build(), snapshots, self)

    def delete_snapshots(self, snapshots_data):
        async def delete(environment, snapshot_names):
            return self.cleanup(env=environment, keep=0, dryrun=False, _id={"$in": snapshot_names})

        def done(f):
            try:
                # just consume the result to raise exception
                # if there were an error... (what an api...)
                f.result()
                logging.info("success", extra={"notify": True})
            except Exception as e:
                logging.exception("failed: %s" % e, extra={"notify": True})

        jobs = []
        try:
            for environment, snapshot_names in snapshots_data.items():
                job = self.job_manager.submit(partial(delete, environment, snapshot_names))
                jobs.append(job)
            tasks = asyncio.gather(*jobs)
            tasks.add_done_callback(done)
        except Exception as ex:
            logging.exception("Error while deleting snapshots. error: %s", ex, extra={"notify": True})
        return jobs



import logging
import os
import time
from datetime import datetime
from enum import Enum
from types import SimpleNamespace

from biothings.utils.common import merge, timesofar


class Stage(Enum):
    READY = 0
    STARTED = 1
    DONE = 2

    def at(self, stage):
        assert self == stage


# IndexJobStateRegistrar CAN be further generalized
# to replace utils.manager.BaseStatusRegisterer


class IndexJobStateRegistrar:
    def __init__(self, collection, build_name, index_name, **context):
        self.collection = collection
        self.build_id = build_name

        self.index_name = index_name
        self.context = context

        self.stage = Stage.READY
        self.t0 = 0

    @staticmethod
    def prune(collection):
        for build in collection.find():
            dirty = False
            for job in build.get("jobs", []):
                if job.get("status") == "in progress":
                    logging.warning("Found stale build '%s', marking index status as 'cancelled'", build["_id"])

                    job["status"] = "cancelled"
                    job.pop("pid", None)
                    dirty = True

            if dirty:
                collection.replace_one({"_id": build["_id"]}, build)

    def started(self, step="index"):
        self.stage.at(Stage.READY)
        self.stage = Stage.STARTED

        self.t0 = time.time()

        job = {
            "step": step,
            "status": "in progress",
            "step_started_at": datetime.now().astimezone(),
            "pid": os.getpid(),
            **self.context,
        }
        self.collection.update(
            {"_id": self.build_id},
            {"$push": {"jobs": job}},
        )

    def failed(self, error):
        def func(job, delta_build):
            job["status"] = "failed"
            job["err"] = str(error)

        self._done(func)

    def succeed(self, result):
        def func(job, delta_build):
            job["status"] = "success"
            if result:
                delta_build["index"] = {self.index_name: result}

        self._done(func)

    def _done(self, func):
        self.stage.at(Stage.STARTED)
        self.stage = Stage.DONE

        build = self.collection.find_one({"_id": self.build_id})
        assert build, "Can't find build document '%s'" % self.build_id

        job = build["jobs"][-1]
        job["time"] = timesofar(self.t0)
        job["time_in_s"] = round(time.time() - self.t0, 0)
        job.pop("pid")

        delta_build = {}
        func(job, delta_build)
        merge(build, delta_build)
        self.collection.replace_one({"_id": build["_id"]}, build)


class PreIndexJSR(IndexJobStateRegistrar):
    def started(self):
        super().started("pre-index")

    def succeed(self, result):
        # no result registration on pre-indexing step.
        # --------------------------------------------
        # registration indicates the creation of
        # the index on the elasticsearch server.
        # thus failure at the post-index stage means
        # registration of the index state up until the
        # indexing step, but success at the pre-index
        # stage suggests no index created and thus
        # no registration at all.
        super().succeed({})


class MainIndexJSR(IndexJobStateRegistrar):
    def started(self):
        super().started("index")


class PostIndexJSR(IndexJobStateRegistrar):
    def started(self):
        super().started("post-index")


# TESTS OUTDATED


def test_registrar():
    from pymongo import MongoClient

    indexer = SimpleNamespace(
        mongo_collection_name="mynews_202012280220_vsdevjdk",  # must exists in DB
        es_client_args=dict(hosts="localhost:9200"),
        es_index_name="__index_name__",
        logfile="/log/file",
        conf_name="bc_news",
        env_name="dev",
    )
    collection = MongoClient().biothings.src_build
    IndexJobStateRegistrar.prune(collection)

    # ----------
    #  round 1
    # ----------

    job = MainIndexJSR(indexer, collection)

    input()
    job.started()
    input()
    job.failed("MockErrorA")
    input()
    try:
        job.succeed()
    except Exception as exc:
        print(exc)

    # ----------
    #  round 2
    # ----------

    job = MainIndexJSR(indexer, collection)

    input()
    job.started()
    input()
    job.succeed(index={"__index_name__": {"count": "99"}})

    # ----------
    #  round 3
    # ----------

    job = PostIndexJSR(indexer, collection)

    input()
    try:
        job.succeed()
    except Exception as exc:
        print(exc)

    input()
    job.started()

    input()
    job.succeed({"__index_name__": {"additionally": "done"}})


if __name__ == "__main__":
    test_registrar()



from collections import UserDict

__all__ = ["DEFAULT_INDEX_SETTINGS", "DEFAULT_INDEX_MAPPINGS", "IndexMappings", "IndexSettings"]


# the hub may create an "all" field for all fields
# we want to enable query string query by default

DEFAULT_INDEX_SETTINGS = {
    "query": {"default_field": "_id,all"},
    "codec": "best_compression",
    "analysis": {
        "analyzer": {
            # Sebastian's Note:
            # soon deprecated in favor of keyword_lowercase_normalizer
            "string_lowercase": {
                "tokenizer": "keyword",
                "filter": "lowercase",
            },
            "whitespace_lowercase": {
                "tokenizer": "whitespace",
                "filter": "lowercase",
            },
        },
        "normalizer": {
            "keyword_lowercase_normalizer": {
                "filter": ["lowercase"],
                "type": "custom",
                "char_filter": [],
            },
        },
    },
}

DEFAULT_INDEX_MAPPINGS = {
    "dynamic": "false",
    "properties": {"all": {"type": "text"}},
}


class _IndexPayload(UserDict):
    async def finalize(self, client):
        """Generate the ES payload format of the corresponding entities
        originally in Hub representation. May require querying the ES client
        for certain metadata to determine the compatible data format."""


class IndexMappings(_IndexPayload):
    async def finalize(self, client):
        version = int((await client.info())["version"]["number"].split(".")[0])
        if version < 7:  # inprecise
            doc_type = self.pop("__hub_doc_type", "doc")
            return {doc_type: dict(self)}
        else:
            self.pop("__hub_doc_type", None)
        return dict(self)


class IndexSettings(_IndexPayload):
    async def finalize(self, client):
        return {"index": dict(self)}


def test_01():
    import asyncio

    from elasticsearch import AsyncElasticsearch

    client = AsyncElasticsearch()
    mappings = IndexMappings(DEFAULT_INDEX_MAPPINGS)
    mappings["dynamic"] = True

    async def finalize_mapping():
        print(await mappings.finalize(client))
        await client.close()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(finalize_mapping())


def test_02():
    import asyncio

    settings = IndexSettings(DEFAULT_INDEX_SETTINGS)
    settings["codec"] = "meow"

    async def finalize_settings():
        print(await settings.finalize(None))

    loop = asyncio.get_event_loop()
    loop.run_until_complete(finalize_settings())


if __name__ == "__main__":
    test_02()



########################################
# APP-SPECIFIC CONFIGURATION VARIABLES #
########################################
# The following variables should or must be defined in your
# own application. Create a config.py file, import that config_common
# file as:
#
#   from config_hub import *
#
# then define the following variables to fit your needs. You can also override any
# any other variables in this file as required. Variables defined as ValueError() exceptions
# *must* be defined
#

import os

from biothings.utils.loggers import setup_default_log

# * 7. Hub Internals *#
# Define hostname for source database
DATA_SRC_SERVER = "localhost"
# Define port for source database
DATA_SRC_PORT = 27017
# Define name for source database
DATA_SRC_DATABASE = "testhub_src"
# Define username for source database connection (or None if not needed)
DATA_SRC_SERVER_USERNAME = ""
# Define password for source database connection (or None if not needed)
DATA_SRC_SERVER_PASSWORD = ""

# Target (merged collection) database connection
# Define hostname for target database (merged collections)
DATA_TARGET_SERVER = "localhost"
# Define port for target database (merged collections)
DATA_TARGET_PORT = 27017
# Define name for target database (merged collections)
DATA_TARGET_DATABASE = "testhub"
# Define username for target database connection (or None if not needed)
DATA_TARGET_SERVER_USERNAME = ""
# Define password for target database connection (or None if not needed)
DATA_TARGET_SERVER_PASSWORD = ""

# Define Hub DB connection
# Internal backend. Default to mongodb
# For now, other options are: mongodb, sqlite3, elasticsearch
HUB_DB_BACKEND = {
    "module": "biothings.utils.es",
    "host": "localhost",
}

# Hub environment (like, prod, dev, ...)
# Used to generate remote metadata file, like "latest.json", "versions.json"
# If non-empty, this constant will be used to generate those url, as a prefix
# with "-" between. So, if "dev", we'll have "dev-latest.json", etc...
# "" means production
HUB_ENV = ""

# * 2. Datasources *#
# List of package paths for active datasources
ACTIVE_DATASOURCES = []

# * 3. Folders *#
# Path to a folder to store all downloaded files, logs, caches, etc...
DATA_ARCHIVE_ROOT = "/tmp/testhub/datasources"

# cached data (it None, caches won't be used at all)
CACHE_FOLDER = None

# Path to a folder to store all 3rd party parsers, dumpers, etc...
DATA_PLUGIN_FOLDER = "/tmp/testhub/plugins"

# Path to folder containing diff files
# Usually inside DATA_ARCHIVE_ROOT
DIFF_PATH = os.path.join(DATA_ARCHIVE_ROOT, "diff")

# Path to folder containing release note files
# Usually inside DATA_ARCHIVE_ROOT
RELEASE_PATH = os.path.join(DATA_ARCHIVE_ROOT, "release")

# Define path to folder which will contain log files
# Usually inside DATA_ARCHIVE_ROOT
LOG_FOLDER = os.path.join(DATA_ARCHIVE_ROOT, "logs")

# Provide a default hub logger instance (use setup_default_log(name,log_folder)
logger = setup_default_log("hub", LOG_FOLDER)

ES_HOST = "localhost:9200"  # optional
ES_INDICES = {"dev": "main_build_configuration"}
ANNOTATION_DEFAULT_SCOPES = ["_id", "symbol"]



import math


class Schedule:
    def __init__(self, total, batch_size):
        self._batch_size = batch_size
        self._state = ""

        self.total = total
        self.scheduled = 0
        self.finished = 0

    @property
    def _batch(self):
        return math.ceil(self.scheduled / self._batch_size)

    @property
    def _batches(self):
        return math.ceil(self.total / self._batch_size)

    @property
    def _percentage(self):
        _percentage = self.scheduled / self.total * 100
        return "%.1f%%" % _percentage

    def suffix(self, string):
        return " ".join((string, "#%d/%d %s" % (self._batch, self._batches, self._percentage)))

    def completed(self):
        if self.finished != self.total:
            raise ValueError(self.finished, self.total)

    def __iter__(self):
        return self

    def __next__(self):
        if self.scheduled >= self.total:
            self._state = "pending, waiting for completion,"
            raise StopIteration()
        self.scheduled += self._batch_size
        if self.scheduled > self.total:
            self.scheduled = self.total
        self._state = self.suffix("running, on batch") + ","
        return self._batch

    def __str__(self):
        return " ".join(
            f"""
            <Schedule {"done" if self.finished >= self.total else self._state}
                total={self.total} scheduled={self.scheduled} finished={self.finished}>
            """.split()
        )


def test_01():
    schedule = Schedule(100, 10)
    for batch in schedule:
        print(batch)
        print(schedule)


def test_02():
    schedule = Schedule(25, 10)
    for batch in schedule:
        print(batch)
        print(schedule)
        print(schedule.suffix("Task"))


def test_03():
    schedule = Schedule(0, 10)
    for batch in schedule:
        print(batch)
        print(schedule)
        print(schedule.suffix("Task"))


def test_04():
    schedule = Schedule(1, 10)
    for batch in schedule:
        print(batch)
        print(schedule)
        print(schedule.suffix("Task"))


if __name__ == "__main__":
    test_02()
